///////////////////////////////////////////////////////////////////////////////
// Simple shader for rendering a GBuffer
///////////////////////////////////////////////////////////////////////////////
shader gbuffer_vertex
{
	#version 400 compatibility
	in vec3 position;
	in vec3 normalIn;
	in vec3 tangentIn;
	in vec3 uvIn; 

	out vec3 world_space_position; 
	out vec3 world_space_normal; 

	uniform mat4 modelMatrix;

	void main()
	{
		world_space_position = (modelMatrix * vec4(position, 1.0)).xyz; 
		world_space_normal	 = (modelMatrix * vec4(normalIn, 0.0)).xyz;
		gl_Position = gl_ModelViewProjectionMatrix * vec4(world_space_position, 1.0); 
	}
}

shader gbuffer_fragment
{
	#version 400 compatibility
	in vec3 world_space_position; 
	in vec3 world_space_normal;
	uniform vec3 diffuse_reflectance; 
	void main()
	{
		gl_FragData[0] = vec4(world_space_position, 1.0); 
		gl_FragData[1].xyz = world_space_normal; 
		gl_FragData[2].xyz = diffuse_reflectance; 
	}
}

///////////////////////////////////////////////////////////////////////////////
// Simple shader for shading using a GBuffer
///////////////////////////////////////////////////////////////////////////////
shader deferred_shading_vertex
{
	#version 400 compatibility
	void main()
	{
		gl_Position = ftransform(); 
	}
}

shader deferred_shading_fragment
{
	#version 400 compatibility
	#define M_PI 3.141592
	uniform sampler2D position_texture;
	uniform sampler2D normal_texture; 
	uniform sampler2D diffuse_reflectance_texture; 
	uniform vec3 light_position; 
	uniform vec3 light_direction; 
	uniform vec3 light_radiance; 
	uniform float light_radius; 

	uniform bool			light_has_shadowmap;
	uniform bool			light_has_compressed_shadow; 
	uniform sampler2DShadow light_shadowmap_sampler; 
	uniform mat4			light_shadowmap_mvp; 
	uniform sampler2D		light_compressed_shadow_result_texture; 

	void main()
	{
		vec3 position = texelFetch(position_texture, ivec2(gl_FragCoord.xy), 0).xyz; 
		vec3 normal = texelFetch(normal_texture, ivec2(gl_FragCoord.xy), 0).xyz; 
		vec3 diffuse_reflectance = texelFetch(diffuse_reflectance_texture, ivec2(gl_FragCoord.xy), 0).xyz; 
		///////////////////////////////////////////////////////////////////
		// Calculate radiance from light
		///////////////////////////////////////////////////////////////////
		vec3 lightSampleLe, wi; 
		{
			vec3 lightSample = light_position; 
			float distSquared = length(lightSample - position) * length(lightSample - position); 
			float lightArea = light_radius * light_radius * M_PI; 
			wi = normalize(lightSample - position); 
			lightSampleLe = vec3(dot(-wi, light_direction) < 0.0 ? vec3(0.0) : light_radiance); 
			float lightPdf = distSquared / (abs(dot(light_direction, -wi)) * lightArea);
			lightSampleLe = lightSampleLe * (1.0/lightPdf); 
		}

		if(light_has_compressed_shadow) {
			float compressedShadowResult = texelFetch(light_compressed_shadow_result_texture, ivec2(gl_FragCoord.xy), 0).x; 
			lightSampleLe *= compressedShadowResult; 
		}
		else if(light_has_shadowmap) {
			vec4 lightMvpPosition = light_shadowmap_mvp * vec4(position, 1.0); 
			lightMvpPosition *= 1.0 / lightMvpPosition.w; 
			vec3 smCoord = (lightMvpPosition.xyz + vec3(1.0)) * 0.5; 
			float shadow = texture(light_shadowmap_sampler, smCoord.xyz); 
			lightSampleLe *= shadow; 
		}
		
		vec3 diffuse = diffuse_reflectance * (1.0f / M_PI); 
		float costerm = max(0.0, dot(wi, normal));
		gl_FragColor.xyz = lightSampleLe * diffuse * costerm;
		gl_FragColor.xyz = pow(abs(gl_FragColor.xyz), vec3(1.0/ 2.2));
	}
}



///////////////////////////////////////////////////////////////////////////////
// Simple constant color shader 
///////////////////////////////////////////////////////////////////////////////
shader simpleVertex
{
	#version 400 compatibility
	out vec3 color; 
	void main()
	{
		color = gl_Color.xyz; 
		gl_Position = ftransform();
	}
}

shader simpleFragment
{
	#version 400 compatibility
	in vec3 color; 
	void main()
	{
		gl_FragColor.xyz = color; 
	}
}
