#include "CPVSBuilder.h"
#include <deque>
#include <omp.h>
#include <sstream>
#include <GL/freeglut.h>
#include "CPVSBuilder_ispc.h"

using namespace std; 
using namespace chag; 

const chag::float3  CPVSBuilder::corners[] = {
		{ -1.0, -1.0, -1.0 },
		{ 1.0, -1.0, -1.0 },
		{ -1.0, 1.0, -1.0 },
		{ 1.0, 1.0, -1.0 },
		{ -1.0, -1.0, 1.0 },
		{ 1.0, -1.0, 1.0 },
		{ -1.0, 1.0, 1.0 },
		{ 1.0, 1.0, 1.0 }
};

///////////////////////////////////////////////////////////////////////////////
// Constructor. Set up GLSL shaders.
///////////////////////////////////////////////////////////////////////////////
CPVSBuilder::CPVSBuilder(void)
{
	m_program_manager = new GLSLProgramManager("CPVSBuilder.glsl"); 
	m_first_layer_program = m_program_manager->getProgramObject("firstLayerVertex", "firstLayerFragment");
	m_second_layer_program = m_program_manager->getProgramObject("secondLayerVertex", "secondLayerFragment");
}

CPVSBuilder::~CPVSBuilder(void)
{
}

///////////////////////////////////////////////////////////////////////////////
// Create a min/max hierarchy for a depth map (on host)
///////////////////////////////////////////////////////////////////////////////
void CPVSBuilder::minMaxMipMap(float *shadowMap, float **shadowMapMin, float **shadowMapMax, const int levels)
{
	for(int l=levels-2; l>=0; l--) {
		int levelResolution = (1<<l); 
		for(int y=0; y<levelResolution; y++) {
			for(int x=0; x<levelResolution; x++) {
				float minA = shadowMapMin[l+1][(y*2+0)*levelResolution*2 + (x*2 + 0)];
				float minB = shadowMapMin[l+1][(y*2+0)*levelResolution*2 + (x*2 + 1)]; 
				float minC = shadowMapMin[l+1][(y*2+1)*levelResolution*2 + (x*2 + 0)]; 
				float minD = shadowMapMin[l+1][(y*2+1)*levelResolution*2 + (x*2 + 1)];
				shadowMapMin[l][y*levelResolution + x] = min(min(minA, minB), min(minC, minD));
				shadowMapMax[l][y*levelResolution + x] = max(
					max(shadowMapMax[l+1][(y*2+0)*levelResolution*2 + (x*2 + 0)], shadowMapMax[l+1][(y*2+0)*levelResolution*2 + (x*2 + 1)]),
					max(shadowMapMax[l+1][(y*2+1)*levelResolution*2 + (x*2 + 0)], shadowMapMax[l+1][(y*2+1)*levelResolution*2 + (x*2 + 1)])
					);
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Generate the min-max hierarchy for the shadow-map
///////////////////////////////////////////////////////////////////////////////
void CPVSBuilder::generateFirstLayerMinMaxHierarchy(float ***shadowMapMin, float ***shadowMapMax, int subtreeLevels, int subtreeResolution,
													Scene *theScene)
{
	// Allocate host side shadow map and minmax hierarchy
	*shadowMapMin = new float* [subtreeLevels]; 
	*shadowMapMax = new float* [subtreeLevels]; 
	for(int l=0; l<subtreeLevels; l++) {
		int size = (1<<l)*(1<<l); 
		(*shadowMapMin)[l] = new float [size]; 
		if(l<subtreeLevels-1) (*shadowMapMax)[l] = new float [size]; 
		else (*shadowMapMax)[l] = (*shadowMapMin)[l]; 
	}
	// Allocate shadow map on GPU
	GLuint framebuffer, texture; 
	glGenFramebuffers(1, &framebuffer); 
	glGenTextures(1, &texture); 
	glBindTexture(GL_TEXTURE_2D, texture); 
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RG32F, subtreeResolution, subtreeResolution, 0, GL_RG, GL_FLOAT, NULL); 
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer); 
	glFramebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture, 0);
	GLenum drawbuffers[] = {GL_COLOR_ATTACHMENT0};
	glDrawBuffers(1, drawbuffers);

	{ // Generate closest layer shadow map
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer); 
		glBlendFuncSeparate(GL_ONE, GL_ONE, GL_ZERO, GL_ONE);  
		glBlendEquationSeparate(GL_MIN, GL_FUNC_ADD);
		glClearColor(2.0f, -1.0f, -1.0f, -1.0f);
		glClear(GL_COLOR_BUFFER_BIT); 
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		m_first_layer_program->enable(); 
		theScene->submitModels(m_first_layer_program);
		m_first_layer_program->disable(); 
		///////////////////////////////////////////////////////////////////
		// Copy to host and minmax-mipmap this shadowmap
		///////////////////////////////////////////////////////////////////
		glReadPixels(0, 0, subtreeResolution, subtreeResolution, GL_RED, GL_FLOAT, (*shadowMapMin)[subtreeLevels - 1]); 
		minMaxMipMap((*shadowMapMin)[subtreeLevels - 1], (*shadowMapMin), (*shadowMapMax), subtreeLevels); 
	}
	// Clean up gpu memory
	glDeleteTextures(1, &texture); 
	glDeleteFramebuffers(1, &framebuffer); 
}

///////////////////////////////////////////////////////////////////////////////
// Generate the min-max hierarchy for the "second-layer" map. 
///////////////////////////////////////////////////////////////////////////////
void CPVSBuilder::generateSecondLayerMinMaxHierarchy(float ***sndLayerMapMin, float ***sndLayerMapMax, int subtreeLevels, int subtreeResolution,
													Scene *theScene)
{
	// Allocate host side second layer map and minmax hierarchy
	*sndLayerMapMin = new float* [subtreeLevels]; 
	*sndLayerMapMax = new float* [subtreeLevels]; 
	for(int l=0; l<subtreeLevels; l++) {
		int size = (1<<l)*(1<<l); 
		(*sndLayerMapMin)[l] = new float [size]; 
		if(l<subtreeLevels-1) (*sndLayerMapMax)[l] = new float [size]; 
		else (*sndLayerMapMax)[l] = (*sndLayerMapMin)[l]; 
	}
	// Allocate GPU side buffers for finding second layer maps
	GLuint framebuffer[2], texture[2], depth[2];
	glGenFramebuffers(2, framebuffer);
	glGenTextures(2, texture); 
	glGenTextures(2, depth); 
	for(int i=0; i<2; i++)
	{
		glBindTexture(GL_TEXTURE_2D, texture[i]); 
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RG32F, subtreeResolution, subtreeResolution, 0, GL_RG, GL_FLOAT, NULL); 
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glBindTexture(GL_TEXTURE_2D, depth[i]); 
		glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT32F, subtreeResolution, subtreeResolution, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL); 
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[i]); 
		glFramebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture[i], 0);
		glFramebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, depth[i], 0);
		GLenum drawbuffers[] = {GL_COLOR_ATTACHMENT0};
		glDrawBuffers(1, drawbuffers);
	}
	
	{ // Generate second layer shadow map
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[0]); 
		glClearColor(2.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[1]); 
		glClearColor(2.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
		GLuint samplesPassedQuery; 
		glGenQueries(1, &samplesPassedQuery); 
		GLuint sampleCount; 
		int pass = 0; 
		do {
			GLuint &current_framebuffer = framebuffer[pass%2]; 
			GLuint &prev_framebuffer = framebuffer[(pass+1)%2]; 
			GLuint &current_texture = texture[pass%2]; 
			GLuint &prev_texture = texture[(pass+1)%2]; 

			glBindFramebuffer(GL_READ_FRAMEBUFFER, prev_framebuffer);
			glBindFramebuffer(GL_DRAW_FRAMEBUFFER, current_framebuffer);
			glBlitFramebuffer(0, 0, subtreeResolution, subtreeResolution, 0, 0, subtreeResolution, subtreeResolution, GL_COLOR_BUFFER_BIT, GL_NEAREST); 
			glBindFramebuffer(GL_FRAMEBUFFER, current_framebuffer);
			glClearDepth(1.0f); 
			glClear(GL_DEPTH_BUFFER_BIT); 
			glDisable(GL_CULL_FACE); 
			glDepthFunc(GL_LESS);
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_BLEND); 
					
			glBeginQuery(GL_SAMPLES_PASSED, samplesPassedQuery);
			m_second_layer_program->enable(); 
			m_second_layer_program->setUniform("pass", pass); 
			m_second_layer_program->bindTexture("txt", prev_texture, GL_TEXTURE_2D, 8);
			theScene->submitModels(m_second_layer_program);
			m_second_layer_program->disable(); 
			glEndQuery(GL_SAMPLES_PASSED); 

			GLint available; 
			do {
				glGetQueryObjectiv(samplesPassedQuery, GL_QUERY_RESULT_AVAILABLE, &available);
			} while (!available);

			glGetQueryObjectuiv(samplesPassedQuery, GL_QUERY_RESULT, &sampleCount);
			pass += 1; 
		} while(sampleCount > 0); 
		glReadPixels(0, 0, subtreeResolution, subtreeResolution, GL_RED, GL_FLOAT, (*sndLayerMapMin)[subtreeLevels-1]); 
		minMaxMipMap((*sndLayerMapMin)[subtreeLevels - 1], (*sndLayerMapMin), (*sndLayerMapMax), subtreeLevels); 
	}
	glDeleteTextures(2, texture); 
	glDeleteTextures(2, depth); 
	glDeleteFramebuffers(2, framebuffer); 
}

///////////////////////////////////////////////////////////////////////////////
// Helper. Set the value of an element in a 16bit childmask. 
///////////////////////////////////////////////////////////////////////////////
void CPVSBuilder::setChild(int i, CPVSBuilder::isectResult r, uint16_t &bitmask) {
	uint16_t val = r << (i*2);
	bitmask &= ~(3 << (i*2)); 
	bitmask |= val; 
};

///////////////////////////////////////////////////////////////////////////////
// Helper. Get the value of an element in a 16bit childmask. 
///////////////////////////////////////////////////////////////////////////////
CPVSBuilder::isectResult CPVSBuilder::getChild(int i, const uint16_t bitmask) {
	return isectResult(0x3 & (bitmask >> (i*2))); 
};

///////////////////////////////////////////////////////////////////////////////
// Update a node at the final levels of the DAG, where the node points to
// eight flat subnodes (8x8 each) at ascending depth. 
///////////////////////////////////////////////////////////////////////////////
CPVSBuilder::NodeUpdateResult CPVSBuilder::updateWithShadowMap_leaf(float **shadowMapMin,
										  float **shadowMapMax, float **sndLayerMapMin, 
										  float **sndLayerMapMax, 
										  const chag::float3 &nodeCenter, 
										  const chag::float3 &nodeRadius, const int level, const int maxLevels,
										  CompactResult & compact_result)
{
	int finalResolution = 1 << (level + 3);
	int2 sliceCoord = make_vector(int((((nodeCenter.x - nodeRadius.x) + 1.0f)/2.0f) * finalResolution), 
								  int((((nodeCenter.y - nodeRadius.y) + 1.0f)/2.0f) * finalResolution));
	float sliceZStart = nodeCenter.z - nodeRadius.z + (nodeRadius.z/8.0f);
	float sliceSize = nodeRadius.z / 4.0f; 

	///////////////////////////////////////////////////////////////////////////
	// Find out if anything happens with the shadow map in this range, 
	// otherwise all leaves will be the same
	///////////////////////////////////////////////////////////////////////////
	bool changes = false; 
	for(int y=0; y<8; y++) {
		for(int x=0; x<8; x++) {
			float shadowMapZ = shadowMapMin[level+3][(sliceCoord.y+y)*finalResolution+(sliceCoord.x+x)]; 
			if(shadowMapZ > (nodeCenter.z - nodeRadius.z) && shadowMapZ < (nodeCenter.z + nodeRadius.z)) {
				changes = true; 
				break;
			}
		}
		if(changes) break;
	}
	///////////////////////////////////////////////////////////////////////////
	// Run through all child-nodes and decide whether they are completely
	// visible, completely in shadow, completely in an undefined region or 
	// whether we need to evaluate visibility for each final voxel.
	///////////////////////////////////////////////////////////////////////////
	NodeUpdateResult child_result[8]; 
	uint64_t leafmask[8]; 
	uint16_t childmask = 0; 
	for(int i=0; i<8; i++) {
		child_result[i].index = 0; 
		leafmask[i] = 0; 
		if(i>0 && !changes) {
			setChild(i, getChild(0, childmask), childmask); 
			child_result[i] = child_result[0]; 
			leafmask[i] = leafmask[0]; 
			continue; 
		}
		float3 childCenter = make_vector(nodeCenter.x, nodeCenter.y, nodeCenter.z - nodeRadius.z + float(0.5f + i) * sliceSize);
		float3 childNodeRadius = nodeRadius; 
		childNodeRadius.z = (1.0f/4.0f) * nodeRadius.z; 
		Aabb childAabb;
		childAabb.min = childCenter - childNodeRadius;
		childAabb.max = childCenter + childNodeRadius;
		// Note, have to look at last shadowmap level here!
		int levelResolution = (1<<(level));
		int2 coords = make_vector(int(((childCenter.x + 1.0f)/2.0f)*levelResolution), 
								  int(((childCenter.y + 1.0f)/2.0f)*levelResolution)); 
		if(sndLayerMapMax[level][coords.y*levelResolution+coords.x] < childAabb.min.z) setChild(i, NODE_IN_SHADOW, childmask); 
		else if(shadowMapMin[level][coords.y*levelResolution+coords.x] > childAabb.max.z) setChild(i, NODE_VISIBLE, childmask); 
		else if( (shadowMapMax[level][coords.y*levelResolution+coords.x] < childAabb.min.z) &&
				 (sndLayerMapMin[level][coords.y*levelResolution+coords.x] > childAabb.max.z)) 
			setChild(i, NODE_UNDEFINED, childmask); 
		else if(sndLayerMapMin[level][coords.y*levelResolution+coords.x] > childAabb.max.z) setChild(i, NODE_VISIBLE, childmask); 
		else {
			setChild(i, NODE_INTERSECTS, childmask); 
			{ // Perform the update on the FlatLeaf here				
				bool undefined = ispc::updateWithShadowMap_leaf(shadowMapMin[level+3], sndLayerMapMax[level+3], 
									sliceCoord.x, sliceCoord.y, sliceZStart + float(i)*sliceSize, level+1, (int *)&leafmask[i]);
				if(undefined) {
					setChild(i, NODE_UNDEFINED, childmask); 
					leafmask[i] = 0; 
				}
				else {
					if(leafmask[i] == 0x0) {
						setChild(i, NODE_VISIBLE, childmask); 
					}
					else if(leafmask[i] == 0xFFFFFFFFFFFFFFFFULL) {
						setChild(i, NODE_IN_SHADOW, childmask); 
					}
					else {
						setChild(i, NODE_INTERSECTS, childmask); 
					}
				}
			}
		}
		child_result[i].state = getChild(i, childmask); 
	}

	///////////////////////////////////////////////////////////////////////////
	// If the node contains subnodes that are completely undefined, decide
	// whether they should be set to visible or in shadow. 
	///////////////////////////////////////////////////////////////////////////
	NodeUpdateResult node_result; 
	node_result.index = 0; 
	node_result.state = NODE_INTERSECTS;
	// Remove all undefineds
	bool hasShadowedNodes = false; 
	for(int i=0; i<8; i++) if((getChild(i, childmask) == NODE_INTERSECTS) || (getChild(i, childmask) == NODE_IN_SHADOW)) hasShadowedNodes = true; 
	bool hasLitNodes = false; 
	for(int i=0; i<8; i++) if((getChild(i, childmask) == NODE_INTERSECTS) || (getChild(i, childmask) == NODE_VISIBLE)) hasLitNodes = true; 
	if(!hasShadowedNodes && !hasLitNodes) {
		node_result.state = NODE_UNDEFINED; 
	}
	else {
		for(int i=0; i<8; i++) {
			if((getChild(i, childmask) == NODE_UNDEFINED)) {
				if(hasShadowedNodes) setChild(i, NODE_IN_SHADOW, childmask); 
				else setChild(i, NODE_VISIBLE, childmask); 
			}
		}
		if(childmask == 0) node_result.state = NODE_IN_SHADOW; 
		else if(childmask == 0x5555) node_result.state = NODE_VISIBLE; 
	}

	///////////////////////////////////////////////////////////////////////////
	// If the whole (8x8x8) node is not completely in shadow or visible, 
	// find out if a node just like it has already been inserted into the 
	// hash-map. Otherwise, insert it and move on. 
	///////////////////////////////////////////////////////////////////////////
	if(node_result.state == NODE_INTERSECTS) {
		// Insert all leaf (8x8) nodes
		for(int i=0; i<8; i++) {
			if(getChild(i, childmask) == NODE_INTERSECTS) {
				auto slot = compact_result.leaf_node_hashmap.find(leafmask[i]); 
				bool firstEncounter = (slot == compact_result.leaf_node_hashmap.end());
				if(firstEncounter) {
					child_result[i].index = uint32_t(compact_result.compact_data.size()); 
					compact_result.compact_data.push_back(uint32_t(leafmask[i] & 0xFFFFFFFF)); 
					compact_result.compact_data.push_back(uint32_t(leafmask[i] >> 32)); 
					compact_result.leaf_node_hashmap.emplace(make_pair(leafmask[i], child_result[i].index));
				}
				else {
					child_result[i].index = slot->second; 
				}
			}
		}
		// Insert 8x8x8 node
		CompactInternalNode cnode; 
		cnode.m_childmask = childmask; 
		for(int i=0; i<8; i++) {
			cnode.m_children[i] = child_result[i].index; 
		}
		auto slot = compact_result.internal_node_hashmap.find(cnode);
		bool firstEncounter = (slot == compact_result.internal_node_hashmap.end());
		if(firstEncounter) {
			node_result.index = uint32_t(compact_result.compact_data.size()); 
			compact_result.compact_data.push_back(cnode.m_childmask); 
			for(int i=0; i<8; i++) {
				if(getChild(i, childmask) == NODE_INTERSECTS) {
					compact_result.compact_data.push_back(cnode.m_children[i]); 
				}
			}
			compact_result.internal_node_hashmap.emplace(make_pair(cnode, node_result.index));
		}
		else {
			node_result.index = slot->second; 
		}
	}
	return node_result; 
}

///////////////////////////////////////////////////////////////////////////////
// Update an internal (not leaf) node. Recursively process all subnodes and 
// then insert this node into the hash-map if it does not already exist. 
///////////////////////////////////////////////////////////////////////////////
CPVSBuilder::NodeUpdateResult CPVSBuilder::updateWithShadowMap_internal(float **shadowMapMin, 
											  float **shadowMapMax, float **sndLayerMapMin, 
											  float **sndLayerMapMax, 
											  const chag::float3 &nodeCenter, 
											  const chag::float3 &nodeRadius, const int level, 
											  const int maxLevels,
											  CompactResult & compact_result)
{
	NodeUpdateResult child_result[8]; 
	uint16_t childmask = 0; 
	
	///////////////////////////////////////////////////////////////////////////
	// Run through all child-nodes and decide whether they are completely
	// visible, completely in shadow, completely in an undefined region or 
	// whether we need recursively process them before continuing with this
	// node. 
	///////////////////////////////////////////////////////////////////////////
	for(int i=0; i<8; i++) {
		child_result[i].index = 0; 
		float3 childCenter = nodeCenter + corners[i] * 0.5f * nodeRadius; 
		float3 childNodeRadius = 0.5f * nodeRadius; 
		Aabb childAabb;
		childAabb.min = childCenter - childNodeRadius;
		childAabb.max = childCenter + childNodeRadius;
		int levelResolution = (1<<(level + 1));
		int2 coords = make_vector(int(((childCenter.x + 1.0f)/2.0f)*levelResolution), 
								  int(((childCenter.y + 1.0f)/2.0f)*levelResolution)); 
		if(sndLayerMapMax[level+1][coords.y*levelResolution+coords.x] < childAabb.min.z) setChild(i, NODE_IN_SHADOW, childmask); 
		else if(shadowMapMin[level+1][coords.y*levelResolution+coords.x] > childAabb.max.z) setChild(i, NODE_VISIBLE, childmask); 
		else if( (shadowMapMax[level+1][coords.y*levelResolution+coords.x] < childAabb.min.z) &&
				 (sndLayerMapMin[level+1][coords.y*levelResolution+coords.x] > childAabb.max.z) ) 
			setChild(i, NODE_UNDEFINED, childmask); 
		else if(sndLayerMapMin[level+1][coords.y*levelResolution+coords.x] > childAabb.max.z) setChild(i, NODE_VISIBLE, childmask); 
		else {
			setChild(i, NODE_INTERSECTS, childmask); 
			if(level == (maxLevels - 5)) {
				child_result[i] = updateWithShadowMap_leaf(shadowMapMin, shadowMapMax, 
															  sndLayerMapMin, sndLayerMapMax, 
															  childCenter, childNodeRadius, level+1, 
															  maxLevels, compact_result);
			}
			else { 
				child_result[i] = updateWithShadowMap_internal(shadowMapMin, shadowMapMax, 
															  sndLayerMapMin, sndLayerMapMax, 
															  childCenter, childNodeRadius, level+1, 
															  maxLevels, compact_result);
			} 
			setChild(i, child_result[i].state, childmask); 
		}
	}

	///////////////////////////////////////////////////////////////////////////
	// If the node contains subnodes that are completely undefined, decide
	// whether they should be set to visible or in shadow. 
	///////////////////////////////////////////////////////////////////////////
	NodeUpdateResult node_result; 
	node_result.index = 0; 
	node_result.state = NODE_INTERSECTS;
	bool hasShadowedNodes = false; 
	for(int i=0; i<8; i++) if((getChild(i, childmask) == NODE_INTERSECTS) || (getChild(i, childmask) == NODE_IN_SHADOW)) hasShadowedNodes = true; 
	bool hasLitNodes = false; 
	for(int i=0; i<8; i++) if((getChild(i, childmask) == NODE_INTERSECTS) || (getChild(i, childmask) == NODE_VISIBLE)) hasLitNodes = true; 
	if(!hasShadowedNodes && !hasLitNodes) {
		node_result.state = NODE_UNDEFINED; 
	}
	else {
		for(int i=0; i<8; i++) {
			if((getChild(i, childmask) == NODE_UNDEFINED)) {
				if(hasShadowedNodes) setChild(i, NODE_IN_SHADOW, childmask); 
				else setChild(i, NODE_VISIBLE, childmask); 
			}
		}
		if(childmask == 0) node_result.state = NODE_IN_SHADOW; 
		else if(childmask == 0x5555) node_result.state = NODE_VISIBLE; 
	}

	///////////////////////////////////////////////////////////////////////////
	// Finally check if a node just like this has already been inserted, and 
	// otherwise insert this into the hash-map. Either way, return the index
	// in the compact data-structure. 
	///////////////////////////////////////////////////////////////////////////
	if(node_result.state == NODE_INTERSECTS) {
		// Insert into compact data
		CompactInternalNode cnode; 
		cnode.m_childmask = childmask; 
		for(int i=0; i<8; i++) {
			cnode.m_children[i] = child_result[i].index; 
		}
		auto slot = compact_result.internal_node_hashmap.find(cnode);
		bool firstEncounter = (slot == compact_result.internal_node_hashmap.end());
		if(firstEncounter) {
			node_result.index = uint32_t(compact_result.compact_data.size()); 
			compact_result.compact_data.push_back(cnode.m_childmask); 
			for(int i=0; i<8; i++) {
				if(getChild(i, childmask) == NODE_INTERSECTS) {
					compact_result.compact_data.push_back(cnode.m_children[i]); 
				}
			}
			compact_result.internal_node_hashmap.emplace(make_pair(cnode, node_result.index));
		}
		else {
			node_result.index = slot->second; 
		}
	}
	return node_result; 
}

///////////////////////////////////////////////////////////////////////////////
// Given a model-view-projection matrix and the scene, create a CPVS DAG
///////////////////////////////////////////////////////////////////////////////
void CPVSBuilder::build(Scene *theScene, Scene::ShadowMapShadowGenerator *shadowMapGenerator, 
						int totalLevels, bool useClosedObjects)
{
	///////////////////////////////////////////////////////////////////////////
	// Divide between toptree and subtree sizes, to avoid rendering too large
	// shadow-maps. 
	///////////////////////////////////////////////////////////////////////////
	int subtreeLevels = min(totalLevels - 1, 12); 
	int subtreeResolution = 1 << (subtreeLevels - 1);
	int toptreeLevels = max(1, totalLevels - subtreeLevels + 1); 
	int toptreeResolution = 1 << (toptreeLevels - 1);
	int totalResolution = 1 << (totalLevels-1); 
	m_levels = totalLevels; 
	LOG_INFO("Total: [levels: " << totalLevels << ", resolution: " << totalResolution << "]"); 
	LOG_INFO("Toptree: [levels: " << toptreeLevels << ", resolution: " << toptreeResolution << "]"); 
	LOG_INFO("Subtree: [levels: " << subtreeLevels << ", resolution: " << subtreeResolution << "]"); 

	///////////////////////////////////////////////////////////////////////////
	// For each "tile" of the shadow-map, build the corresponding sub-DAGs. 
	// For each sub-volume calculate the sub-DAG and store a pointer to it's
	// compact representation in a grid. 
	///////////////////////////////////////////////////////////////////////////
	glPushAttrib(GL_ALL_ATTRIB_BITS); 
	view lightView = shadowMapGenerator->getView(); 
	glViewport(0, 0, subtreeResolution, subtreeResolution); 
	// Allocate space for leading root node in compact result
	m_compact_result.compact_data.resize(9); 

	NodeUpdateResult *topgrid = new NodeUpdateResult[toptreeResolution * toptreeResolution * toptreeResolution];
	for(int passY = 0; passY < toptreeResolution; passY++) {
		for(int passX = 0; passX < toptreeResolution; passX++) {
			glFinish(); unsigned int current_time = glutGet(GLUT_ELAPSED_TIME);

			///////////////////////////////////////////////////////////////////
			// Generate projection matrix for current tile
			///////////////////////////////////////////////////////////////////
			lightView.push();
			glMatrixMode(GL_PROJECTION); 
			glLoadIdentity(); 
			float4x4 P = lightView.get_P(); 
			glTranslatef(float((toptreeResolution-1) - passX*2), float((toptreeResolution-1) - passY*2), 0.0f); 
			glScalef(float(toptreeResolution), float(toptreeResolution), 1.0); 
			glMultMatrixf(&P.c1.x);
			glMatrixMode(GL_MODELVIEW); 
			glViewport(0, 0, subtreeResolution, subtreeResolution); 

			///////////////////////////////////////////////////////////////////
			// Render shadowmap min max hierarchies
			///////////////////////////////////////////////////////////////////
			float **shadowMapMin = NULL, **shadowMapMax = NULL;
			generateFirstLayerMinMaxHierarchy(&shadowMapMin, &shadowMapMax, subtreeLevels, subtreeResolution, theScene); 
			// If we don't use closed objects, set second layer to shadow-map and 
			// we can treat all as closed. 
			float **sndLayerMapMin = shadowMapMin, **sndLayerMapMax = shadowMapMax;
			if(useClosedObjects) {
				generateSecondLayerMinMaxHierarchy(&sndLayerMapMin, &sndLayerMapMax, subtreeLevels, subtreeResolution, theScene); 
			}
			lightView.pop();
			glFinish(); 

			///////////////////////////////////////////////////////////////////
			// Process each subtree under current tile
			///////////////////////////////////////////////////////////////////
			for(int passZ = 0; passZ < toptreeResolution; passZ++) {
				float subtreeRootRadius = 1.0f / toptreeResolution;
				NodeUpdateResult subtree_cell; 
				subtree_cell = 
					updateWithShadowMap_internal(shadowMapMin, shadowMapMax, 
												 sndLayerMapMin, sndLayerMapMax,
												 make_vector(0.0f, 0.0f, -1.0f + subtreeRootRadius + 2* passZ*subtreeRootRadius), 
												 make_vector(1.0f, 1.0f, subtreeRootRadius), 0, subtreeLevels,
												 m_compact_result); 
				topgrid[passZ * (toptreeResolution*toptreeResolution) + passY * toptreeResolution + passX] = subtree_cell;
			}
			///////////////////////////////////////////////////////////////////
			// Release shadow map memory
			///////////////////////////////////////////////////////////////////
			for(int l=0; l<subtreeLevels; l++) {
				delete [] shadowMapMin[l]; 
				if(l<subtreeLevels-1) delete [] shadowMapMax[l];
			}
			delete [] shadowMapMin; 
			delete [] shadowMapMax; 
			if(useClosedObjects) {
				for(int l=0; l<subtreeLevels; l++) {
					delete [] sndLayerMapMin[l]; 
					if(l<subtreeLevels-1) delete [] sndLayerMapMax[l];
				}
				delete [] sndLayerMapMin; 
				delete [] sndLayerMapMax; 
			}
			float progress = float(passY*toptreeResolution + passX) / float(toptreeResolution*toptreeResolution);
			LOG_INFO("Processing..." << progress * 100.0f << "% [" << float(m_compact_result.compact_data.size()*4) / (1024.0f*1024.0f) << "MB]"); 
		}
	}
	///////////////////////////////////////////////////////////////////////////
	// Finalize DAG from topgrid
	///////////////////////////////////////////////////////////////////////////
	int src_resolution = toptreeResolution; 
	int dst_resolution = toptreeResolution / 2; 
	NodeUpdateResult *src_grid = topgrid; 
	NodeUpdateResult *dst_grid = new NodeUpdateResult[dst_resolution * dst_resolution * dst_resolution]; 	
	while(src_resolution > 1) {
		for(int Z=0; Z<dst_resolution; Z++) {
			for(int Y=0; Y<dst_resolution; Y++) {
				for(int X=0; X<dst_resolution; X++) {
					NodeUpdateResult child_index[8]; 
					memset(child_index, 0, 8*sizeof(uint32_t)); 
					child_index[0] = src_grid[(Z*2+0)*(src_resolution*src_resolution) + (Y*2+0) * src_resolution + (X*2+0)]; 
					child_index[1] = src_grid[(Z*2+0)*(src_resolution*src_resolution) + (Y*2+0) * src_resolution + (X*2+1)]; 
					child_index[2] = src_grid[(Z*2+0)*(src_resolution*src_resolution) + (Y*2+1) * src_resolution + (X*2+0)]; 
					child_index[3] = src_grid[(Z*2+0)*(src_resolution*src_resolution) + (Y*2+1) * src_resolution + (X*2+1)]; 
					child_index[4] = src_grid[(Z*2+1)*(src_resolution*src_resolution) + (Y*2+0) * src_resolution + (X*2+0)]; 
					child_index[5] = src_grid[(Z*2+1)*(src_resolution*src_resolution) + (Y*2+0) * src_resolution + (X*2+1)]; 
					child_index[6] = src_grid[(Z*2+1)*(src_resolution*src_resolution) + (Y*2+1) * src_resolution + (X*2+0)]; 
					child_index[7] = src_grid[(Z*2+1)*(src_resolution*src_resolution) + (Y*2+1) * src_resolution + (X*2+1)]; 
					uint16_t childmask = 0; 
					// Set up current node childmask
					for(int i=0; i<8; i++) { setChild(i, child_index[i].state, childmask); }
					// Remove all undefineds
					bool hasShadowedNodes = false; 
					for(int i=0; i<8; i++) if((getChild(i, childmask) == NODE_INTERSECTS) || (getChild(i, childmask) == NODE_IN_SHADOW)) hasShadowedNodes = true; 
					bool hasLitNodes = false; 
					for(int i=0; i<8; i++) if((getChild(i, childmask) == NODE_INTERSECTS) || (getChild(i, childmask) == NODE_VISIBLE)) hasLitNodes = true; 
					if(!hasShadowedNodes && !hasLitNodes) {
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].state = NODE_UNDEFINED; 
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].index = 0; 
					}
					else {
						for(int i=0; i<8; i++) {
							if((getChild(i, childmask) == NODE_UNDEFINED)) {
								if(hasShadowedNodes) setChild(i, NODE_IN_SHADOW, childmask); 
								else setChild(i, NODE_VISIBLE, childmask); 
							}
						}
					}
					if(childmask == 0) {
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].state = NODE_IN_SHADOW; 
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].index = 0; 
					}
					else if(childmask == 0x5555) {
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].state = NODE_VISIBLE; 
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].index = 0; 
					}
					else {
						CompactInternalNode cnode; 
						cnode.m_childmask = childmask; 
						for(int i=0; i<8; i++) {
							cnode.m_children[i] = (getChild(i, childmask) == NODE_INTERSECTS) ? child_index[i].index : 0; 
						}
						auto slot = m_compact_result.internal_node_hashmap.find(cnode);
						bool firstEncounter = (slot == m_compact_result.internal_node_hashmap.end());
						uint32_t my_index; 
						if(firstEncounter) {
							my_index = uint32_t(m_compact_result.compact_data.size()); 
							m_compact_result.compact_data.push_back(cnode.m_childmask); 
							for(int i=0; i<8; i++) {
								if(getChild(i, childmask) == NODE_INTERSECTS) {
									m_compact_result.compact_data.push_back(cnode.m_children[i]); 
								}
							}
							m_compact_result.internal_node_hashmap.emplace(make_pair(cnode, my_index));
						}
						else {
							my_index = slot->second; 
						}
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].state = NODE_INTERSECTS; 
						dst_grid[Z*(dst_resolution*dst_resolution)+Y*dst_resolution+X].index = my_index; 
					}
				}
			}
		}
		NodeUpdateResult *temp = src_grid; 
		src_grid = dst_grid; 
		dst_grid = temp; 
		src_resolution /= 2; 
		dst_resolution /= 2; 
	}
	///////////////////////////////////////////////////////////////////////////
	// Copy root node from end to start of compact data
	///////////////////////////////////////////////////////////////////////////
	uint32_t root_index = src_grid[0].index; 
	int c = 0; 
	while(root_index < m_compact_result.compact_data.size()) {
		m_compact_result.compact_data[c] = m_compact_result.compact_data[root_index]; 
		root_index += 1; 
		c += 1; 
	}

	LOG_INFO("Build complete. Size of final DAG: " << m_compact_result.compact_data.size() * 4 << " bytes."); 
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glPopAttrib(); 
}


void CPVSBuilder::saveCompactData(const std::string &filename)
{
	///////////////////////////////////////////////////////////////////////////
	// Save DAG to disc. 
	///////////////////////////////////////////////////////////////////////////
	ofstream os(filename, ios::binary); 
	if(!os.is_open()) LOG_ERROR("Failed to open outputfile " << filename); 
	os.write((char *)&m_levels, sizeof(unsigned int)); 
	unsigned int s = unsigned int(m_compact_result.compact_data.size()); 
	os.write((char *)&s, sizeof(unsigned int)); 
	for(int i=0; i<m_compact_result.compact_data.size(); i++) {
		os.write((char *)&m_compact_result.compact_data[i], sizeof(unsigned int)); 
	}
	os.close(); 
}
