#pragma once
#include <scene/Scene.h>
#include <vector>
#include <functional>
#include <unordered_map>

class CPVSBuilder
{
private: 
	///////////////////////////////////////////////////////////////////////////
	// Helper classes
	///////////////////////////////////////////////////////////////////////////
	enum isectResult { NODE_IN_SHADOW, NODE_VISIBLE, NODE_INTERSECTS, NODE_UNDEFINED };
	struct CompactInternalNode {
		uint32_t m_childmask;
		uint32_t m_children[8];
		inline bool operator == (const CompactInternalNode & rhs) const {
			if (m_childmask != rhs.m_childmask) return false;
			for (int i = 0; i<8; i++) {
				if (m_children[i] != rhs.m_children[i]) return false;
			}
			return true;
		}
	};
	struct InternalNodeHasher {
		size_t operator()(const CompactInternalNode& node) const {
			std::hash<uint32_t> children_hash;
			std::hash<uint32_t> childmask_hash;
			size_t seed = childmask_hash(node.m_childmask);
			for (int i = 0; i<8; i++) {
				if ((node.m_childmask >> i) & 0x1)
					seed ^= children_hash(node.m_children[i]) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
			}
			return seed;
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// Class that stores the hash-map used for building and the final compact
	// DAG. 
	///////////////////////////////////////////////////////////////////////////
	struct CompactResult
	{
		std::unordered_map<CompactInternalNode, uint32_t, InternalNodeHasher> internal_node_hashmap;
		std::unordered_map<uint64_t, uint32_t> leaf_node_hashmap;
		std::vector<uint32_t> compact_data;
	} m_compact_result;

	GLSLProgramManager	*m_program_manager;
	GLSLProgramObject	*m_second_layer_program;
	GLSLProgramObject	*m_first_layer_program;
	uint32_t			 m_levels;

	struct NodeUpdateResult { uint32_t index; isectResult state; };
	static const chag::float3 corners[]; 

	void minMaxMipMap(float *shadowMap, float **shadowMapMin, float **shadowMapMax, const int levels); 
	void generateFirstLayerMinMaxHierarchy(float ***shadowMapMin, float ***shadowMapMax, int subtreeLevels, int subtreeResolution, Scene *theScene);
	void generateSecondLayerMinMaxHierarchy(float ***sndLayerMapMin, float ***sndLayerMapMax, int subtreeLevels, int subtreeResolution, Scene *theScene);
	void setChild(int i, isectResult r, uint16_t &bitmask); 
	isectResult getChild(int i, const uint16_t bitmask);
	NodeUpdateResult updateWithShadowMap_leaf(float **shadowMapMin, float **shadowMapMax,
											  float **sndLayerMapMin, float **sndLayerMapMax,
											  const chag::float3 &nodeCenter,
											  const chag::float3 &nodeRadius, const int level,
											  const int maxLevels, CompactResult & compact_result);
	NodeUpdateResult updateWithShadowMap_internal(float **shadowMapMin, float **shadowMapMax,
												  float **sndLayerMapMin, float **sndLayerMapMax,
												  const chag::float3 &nodeCenter,
												  const chag::float3 &nodeRadius, const int level,
												  const int maxLevels, CompactResult & compact_result);

public:
	CPVSBuilder(void);
	~CPVSBuilder(void);
	void build(Scene *scene, Scene::ShadowMapShadowGenerator *shadowMapGenerator, int totalLevels, bool useClosedObjects);
	void saveCompactData(const std::string &filename);
};


