#include "CompressedShadow.h"
#include <cuda_gl_interop.h>
#include <utils/errorCheck.cuh>
#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// The texture which will contain visibility values for each pixel
///////////////////////////////////////////////////////////////////////////////
surface<void, cudaSurfaceType2D> shadowSurface; 
cudaGraphicsResource *CompressedShadow::m_shadowResource; 
cudaArray			 *CompressedShadow::m_shadowArray;
///////////////////////////////////////////////////////////////////////////////
// The texture containing each viewsamples coordinates in the light's 
// Notmalized Device Coordinates
///////////////////////////////////////////////////////////////////////////////
texture<float4, cudaTextureType2D> positionTexture;
cudaGraphicsResource *CompressedShadow::m_positionResource; 
cudaArray			 *CompressedShadow::m_positionArray;
texture<float4, cudaTextureType2D> normalTexture;
cudaGraphicsResource *CompressedShadow::m_normalResource; 
cudaArray			 *CompressedShadow::m_normalArray;
///////////////////////////////////////////////////////////////////////////////
// Width and height of gbuffer
///////////////////////////////////////////////////////////////////////////////
unsigned int CompressedShadow::m_width; 
unsigned int CompressedShadow::m_height; 

///////////////////////////////////////////////////////////////////////////////
// Setup and resource managment
///////////////////////////////////////////////////////////////////////////////
CompressedShadow::CompressedShadow()
{
	m_bias = 0.1f;
	m_filter_size = 9; 
}

void CompressedShadow::unregisterTextures()
{
	if(m_shadowResource != NULL) {
		_CHECKED_ cudaGraphicsUnregisterResource(m_shadowResource);
		m_shadowResource = NULL; 
	} 
	if(m_positionResource != NULL) {
		_CHECKED_ cudaGraphicsUnregisterResource(m_positionResource);
		m_positionResource = NULL; 
	} 
	if(m_normalResource != NULL) {
		_CHECKED_ cudaGraphicsUnregisterResource(m_normalResource);
		m_normalResource = NULL; 
	} 
}
void CompressedShadow::registerTextures(GLuint shadowTexture, GLuint positionTexture, GLuint normalTexture)
{
	_CHECKED_ cudaGraphicsGLRegisterImage(&m_shadowResource, shadowTexture, GL_TEXTURE_2D, cudaGraphicsRegisterFlagsSurfaceLoadStore | cudaGraphicsRegisterFlagsWriteDiscard);
	_CHECKED_ cudaGraphicsGLRegisterImage(&m_positionResource, positionTexture, GL_TEXTURE_2D, cudaGraphicsRegisterFlagsReadOnly);
	_CHECKED_ cudaGraphicsGLRegisterImage(&m_normalResource, normalTexture, GL_TEXTURE_2D, cudaGraphicsRegisterFlagsReadOnly);
}
void CompressedShadow::mapTextures()
{
	_CHECKED_ cudaGraphicsMapResources(1, &m_shadowResource);
	_CHECKED_ cudaGraphicsSubResourceGetMappedArray( &m_shadowArray, m_shadowResource, 0, 0);
	_CHECKED_ cudaBindSurfaceToArray(shadowSurface, m_shadowArray);

	_CHECKED_ cudaGraphicsMapResources(1, &m_positionResource);
	_CHECKED_ cudaGraphicsSubResourceGetMappedArray( &m_positionArray, m_positionResource, 0, 0);
	_CHECKED_ cudaBindTextureToArray(positionTexture, m_positionArray, cudaCreateChannelDesc<float4>());

	_CHECKED_ cudaGraphicsMapResources(1, &m_normalResource);
	_CHECKED_ cudaGraphicsSubResourceGetMappedArray( &m_normalArray, m_normalResource, 0, 0);
	_CHECKED_ cudaBindTextureToArray(normalTexture, m_normalArray, cudaCreateChannelDesc<float4>());
}
void CompressedShadow::unmapTextures()
{
	_CHECKED_ cudaGraphicsUnmapResources(1, &m_shadowResource, 0);
	_CHECKED_ cudaGraphicsUnmapResources(1, &m_positionResource, 0);
	_CHECKED_ cudaGraphicsUnmapResources(1, &m_normalResource, 0);
}
void CompressedShadow::uploadData() 
{
	cudaMalloc(&m_d_data, m_data_size * sizeof(unsigned int)); 
	cudaMemcpy(m_d_data, m_data, m_data_size * sizeof(unsigned int), cudaMemcpyHostToDevice);
	unsigned int gridSize = (1 << (TOPGRID_LEVELS-1))*(1 << (TOPGRID_LEVELS-1))*(1 << (TOPGRID_LEVELS-1))*sizeof(unsigned int);
	cudaMalloc(&m_d_topGrid, gridSize);
	cudaMemcpy(m_d_topGrid, m_topGrid, gridSize, cudaMemcpyHostToDevice);
}


///////////////////////////////////////////////////////////////////////////////
// Below are a number of helper functions, that should be self-explanatory
///////////////////////////////////////////////////////////////////////////////
__device__ int2 make_int2(int3 v) { return make_int2(v.x,v.y); };
__device__ int2 make_int2(int v) { return make_int2(v,v); };
__device__ float3 make_float3(float v) { return make_float3(v,v,v); };
__device__ int2 inline operator & (int2 a, int2 b) { return make_int2(a.x&b.x, a.y&b.y); };
__device__ uint2 inline operator & (uint2 a, uint2 b) { return make_uint2(a.x&b.x, a.y&b.y); };
__device__ int2 inline operator - (int2 a, int2 b) { return make_int2(a.x-b.x, a.y-b.y); };
__device__ int3 inline operator >> (int3 a, int b) { return make_int3(a.x>>b, a.y>>b, a.z>>b); };
__device__ int2 min(int2 a, int2 b) { return make_int2(min(a.x, b.x), min(a.y, b.y)); }
__device__ int2 max(int2 a, int2 b) { return make_int2(max(a.x, b.x), max(a.y, b.y)); }
__device__ void operator += (float3& a, float3 &b) { a.x += b.x; a.y += b.y; a.z += b.z; }
__device__ void operator *= (float3 &a, float b) { a.x *= b; a.y *= b; a.z *= b; }
__device__ void operator *= (float4& a, float b) { a.x *= b; a.y *= b; a.z *= b; a.w *= b; }
__device__ float3 make_float3(float4 a) { return make_float3(a.x, a.y, a.z); }
__device__ int2 operator + (int2 a, int2 b) { return make_int2(a.x + b.x, a.y + b.y); }
__device__ float3 operator + (float3 a, float3 b) { return make_float3(a.x + b.x, a.y + b.y, a.z + b.z); }
__device__ float3 operator * (float3 a, float b) { return make_float3(a.x * b, a.y * b, a.z * b); }

///////////////////////////////////////////////////////////////////////////////
// Given a point in the lights Normalized Device Coordinates, calculate it's 
// discrete coordinates within a grid. 
///////////////////////////////////////////////////////////////////////////////
__device__ int3 NDCToPath(float3 ndc, int maxlevels)
{
	int resolution = (1 << (maxlevels-1));
	ndc += make_float3(1.0f, 1.0f, 1.0f); 
	ndc *= 0.5f; 
	return make_int3(int(ndc.x * resolution), int(ndc.y * resolution), int(ndc.z * resolution));
}

///////////////////////////////////////////////////////////////////////////////
// Simple matrix - Could be much prettier and should not be here
///////////////////////////////////////////////////////////////////////////////
struct matrix {	float4 c0,c1,c2,c3; };
__device__ float4 matrixmul(const float3 v, const matrix m)
{
	float4 V; 
	V.x = m.c0.x * v.x + m.c1.x * v.y + m.c2.x * v.z + m.c3.x;
	V.y = m.c0.y * v.x + m.c1.y * v.y + m.c2.y * v.z + m.c3.y;
	V.z = m.c0.z * v.x + m.c1.z * v.y + m.c2.z * v.z + m.c3.z;
	V.w = m.c0.w * v.x + m.c1.w * v.y + m.c2.w * v.z + m.c3.w;
	return V;
}
__device__ float4 matrixmul(const float4 v, const matrix m)
{
	float4 V; 
	V.x = m.c0.x * v.x + m.c1.x * v.y + m.c2.x * v.z + m.c3.x * v.w;
	V.y = m.c0.y * v.x + m.c1.y * v.y + m.c2.y * v.z + m.c3.y * v.w;
	V.z = m.c0.z * v.x + m.c1.z * v.y + m.c2.z * v.z + m.c3.z * v.w;
	V.w = m.c0.w * v.x + m.c1.w * v.y + m.c2.w * v.z + m.c3.w * v.w;
	return V;
}

///////////////////////////////////////////////////////////////////////////////
// lastCommonLevel - Calculate the level at which two paths diverge
///////////////////////////////////////////////////////////////////////////////
__device__ unsigned char lastCommonLevel(int3 newPath, int3 oldPath, int maxlevels)
{
	int3 xored = make_int3(newPath.x ^ oldPath.x, newPath.y ^ oldPath.y, newPath.z ^ oldPath.z); 
	int diffBit = __clz(xored.x | xored.y | xored.z);
	return diffBit - (32 - maxlevels) - 1; 
}

///////////////////////////////////////////////////////////////////////////////
// traversalOrder - Written as a function here, but can just as well be a
//                  constant memory lookup table. The returned word defines 
//                  the order in which a number of leaf nodes shall be 
//                  processed, for optimal performance. 
//                
//                  In the returned word, each nibble    Leaf ordering:
//                  points out the leaf that shall be 	 | 8 | 9 | A |
//                  processed, with leaf 0 being the 	 | 4 | 5 | 6 |
//                  lower left leaf in the filter.		 | 0 | 1 | 2 |
// 
//                  In order to fit within a 32-bit word, the first leaf to
//                  visit is implicit, and calculated beforehand. 
///////////////////////////////////////////////////////////////////////////////
__device__ unsigned int traversalOrder(int idx)
{
	switch(idx) {
	// 3x3 leafs covered                StartLeaf:
	case 0: return 0x10452689U << 1; // (A)       
	case 1: return 0x562104A9U << 1; // (8)       
	case 2: return 0x4895A601U << 1; // (2)       
	case 3: return 0x65A98421U << 1; // (0)       
	case 4: return 0x10458926U << 1; // (A)       
	case 5: return 0x5621A904U << 1; // (8)       
	case 6: return 0x489501A6U << 1; // (2)       
	case 7: return 0x65A92184U << 1; // (0)       
	// 3x2 leafs covered
	case 8: return 0x62154U << 1;    // (0)
	case 9: return 0x45621U << 1;    // (0)
	// 2x3 leafs covered
	case 10: return 0x15984U << 1;   // (0) 
	case 11: return 0x98451U << 1;   // (0)
	// 2x2 leafs covered
	case 12: return 0x154U << 1;     // (0)
	case 13: return 0x451U << 1;     // (0)
	// 2x1 leafs covered
	case 14: return 0x1;			 // (0)
	// 1x2 leafs covered
	case 15: return 0x4;			 // (0)
	// 1x1 leafs covered
	case 16: return 0x0;			 // (0)
	}
	return 0; 
}

///////////////////////////////////////////////////////////////////////////////
// shadowLookup - Traverse the dag to find a filtered visibility value. 
///////////////////////////////////////////////////////////////////////////////
template<int LEVELS>
__device__ float shadowLookup(const float3 position,                     // Light NDC coords of viewsample, 
							  const unsigned int * __restrict__ dag, 		 // Device pointer to DAG
							  const unsigned int * __restrict__ topGrid, 	 // Device pointer to top grid
							  int filterWidthAndHeight)					 // Actual filter size (can be dynamic)
{
	///////////////////////////////////////////////////////////////////////////
	// Calculate discrete coordinates (path) of viewsample
	///////////////////////////////////////////////////////////////////////////
	const int3 positionPath = NDCToPath(position, LEVELS);
	///////////////////////////////////////////////////////////////////////////
	// Calculate the discrete coordinates of the (minimum) lower left and 
	// (maxmimum) upper right points on the filter
	///////////////////////////////////////////////////////////////////////////
	const int2 LLPath = make_int2(positionPath) - make_int2(filterWidthAndHeight/2);
	const int2 URPath = LLPath + make_int2(filterWidthAndHeight - 1); 
	///////////////////////////////////////////////////////////////////////////
	// Calculate the two first levels at which these paths diverge
	///////////////////////////////////////////////////////////////////////////
	int divergeLevelX = __clz(LLPath.x^URPath.x)- (32 - LEVELS) - 1;
	int divergeLevelY = __clz(LLPath.y^URPath.y)- (32 - LEVELS) - 1;
	///////////////////////////////////////////////////////////////////////////
	// Calculate the number of leaves the filter covers in x and y 
	// (currently max 3)
	///////////////////////////////////////////////////////////////////////////
	int leavesX = (((URPath.x & (~0x7)) - ((LLPath.x) & (~0x7))) >> 3) + 1; 
	int leavesY = (((URPath.y & (~0x7)) - ((LLPath.y) & (~0x7))) >> 3) + 1;
	///////////////////////////////////////////////////////////////////////////
	// We will now decide the optimal traversal order. First initialize the
	// path to the leaf that we will start from (x,y)
	///////////////////////////////////////////////////////////////////////////
	int lutIdx = 16; // Lookup-table index for the chosen traversal order
	int x = LLPath.x;
	int y = LLPath.y;
	///////////////////////////////////////////////////////////////////////////
	// Depending on which of the crossings are more expensive to cross (x or y)
	// choose whether to take the red or green path in Figure 7
	///////////////////////////////////////////////////////////////////////////
	int expensiveCrossing = (divergeLevelX>divergeLevelY) ? 0 : 1; 
	///////////////////////////////////////////////////////////////////////////
	// If we need to visit 3x3 nodes, choose a traversal   +---+---+
	// order based on the situation and pathColor (see 	   | 1 | 0 |
	// Figure 7c and right). We also need to choose which  +---+---+
	// leaf to start from (not in traversal order because  | 3 | 2 |
	// it doesn't fit in 32 bits).						   +---+---+
	///////////////////////////////////////////////////////////////////////////
	if(leavesX == 3 && leavesY == 3) { 
		int situation = (LLPath.x & 0x8) ? 1 : 0 + (LLPath.y & 0x8) ? 2 : 0; 
		lutIdx = expensiveCrossing * 4 + situation;
		x += (situation&0x1) ? 0 : 2*8;
		y += (situation&0x2) ? 0 : 2*8;
	}
	///////////////////////////////////////////////////////////////////////////
	// Otherwise, if we need to visit 3x2 leafs, or 2x3 leafs, the filter will
	// always touch four leaves from one node (within which the order does not 
	// matter) and two from another, so the order depends only on which is the
	// most expensive crossing.
	///////////////////////////////////////////////////////////////////////////
	else if(leavesX == 3 && leavesY == 2) {
		lutIdx = 8 + (expensiveCrossing ? 1 : 0); 
	}
	else if(leavesX == 2 && leavesY == 3) {
		lutIdx = 10 + (expensiveCrossing ? 1 : 0); 
	}
	///////////////////////////////////////////////////////////////////////////
	// Same for 2x2, see Figure 7a
	///////////////////////////////////////////////////////////////////////////
	else if(leavesX == 2 && leavesY == 2) {
		lutIdx = 12 + (expensiveCrossing ? 1 : 0); 
	}
	///////////////////////////////////////////////////////////////////////////
	// If we touch 1x2, 2x1 or 1x1 leaves, the order does not matter
	///////////////////////////////////////////////////////////////////////////
	else {
		if(leavesX == 2 && leavesY == 1) lutIdx = 14;
		if(leavesX == 1 && leavesY == 2) lutIdx = 15;
		if(leavesX == 1 && leavesY == 1) lutIdx = 16;
	}

	///////////////////////////////////////////////////////////////////////////
	// Now fetch the chosen order from a lookup table 
	///////////////////////////////////////////////////////////////////////////
	unsigned int travOrder = traversalOrder(lutIdx);
	///////////////////////////////////////////////////////////////////////////
	// The two crossings occur at the only two levels to which we will need 
	// to backtrack. We will need stack entries for only these two levels
	///////////////////////////////////////////////////////////////////////////
	const int topBacktrackLevel = min(divergeLevelX, divergeLevelY); 
	const int bottomBacktrackLevel = max(divergeLevelX, divergeLevelY); 
	unsigned int topBacktrackStackEntry = 0;
	unsigned int bottomBacktrackStackEntry = 0;
	///////////////////////////////////////////////////////////////////////////
	// Set up state 
	///////////////////////////////////////////////////////////////////////////
	unsigned int offset = 0;  // Offset to the current node
	unsigned int level = 0;   // The current level
	int3 lastPath;			  // Keep track of the last accessed node (path)
	int finalVisibility = 0;  // Final visibility will be accumulated from all leaves
	///////////////////////////////////////////////////////////////////////////
	// Process all leaves in traversal order 
	///////////////////////////////////////////////////////////////////////////
	bool thereAreLeafnodesLeft = true;
	while(thereAreLeafnodesLeft){
		///////////////////////////////////////////////////////////////////////
		// Path to the current leaf (z is always the same)
		///////////////////////////////////////////////////////////////////////
		const int3 leafPath = make_int3(x, y, positionPath.z);
		///////////////////////////////////////////////////////////////////////
		// Backtrack to the last level that is common to this and the previous 
		// path. Note that the last previous path may not have reached the 
		// common level, so never backtrack further than the last visited
		// level. 
		///////////////////////////////////////////////////////////////////////
		unsigned int lastCommon = min(level, lastCommonLevel(leafPath, lastPath, LEVELS));
		// Never backtrack above the top grid or past the final three levels
		level = max(TOPGRID_LEVELS-1, min(LEVELS - 3, lastCommon));
		// Update offset to our bottom stack entry (or root, if it does not exist yet). 
		if(level<=bottomBacktrackLevel) offset = bottomBacktrackStackEntry;
		// Overwrite with offset to our top stack entry, if the new path diverges
		// that high up. 
		if(level==topBacktrackLevel) offset = topBacktrackStackEntry;

		///////////////////////////////////////////////////////////////////////
		// Now, we will establish the visibility of the child node
		///////////////////////////////////////////////////////////////////////
		enum NodeVisibility { SHADOW = 0, VISIBLE = 1, PARTIAL = 2}; 
		NodeVisibility childVisibility = PARTIAL;

		///////////////////////////////////////////////////////////////////////
		// For the first couple of levels, the offsets to the contained
		// subtrees are stored in a grid. 
		///////////////////////////////////////////////////////////////////////
		if(level == TOPGRID_LEVELS-1) {
			unsigned int res = (1 << (TOPGRID_LEVELS - 1)); 
			int3 topCoords = leafPath >> (LEVELS - TOPGRID_LEVELS); 
			offset = topGrid[topCoords.z * res * res + topCoords.y * res + topCoords.x];
			// The grid stores special values if the entire top-grid--cell is
			// visible or in shadow. 
			if(offset == 0xFFFFFFFF) childVisibility = SHADOW; 
			if(offset == 0xFFFFFFFE) childVisibility = VISIBLE; 
		}

		///////////////////////////////////////////////////////////////////////
		// Traverse down the data structure towards the current leaf node, 
		// until we hit an entirely shadowed or entirely visible node OR
		// if we hit the level ABOVE the leaf nodes (that level is handled
		// separately below)
		///////////////////////////////////////////////////////////////////////
		unsigned int lastOffset = offset;
		unsigned int childMask; 
		while((level < (LEVELS - 4)) && (childVisibility == PARTIAL)) {
			offset = lastOffset;
			// Fetch the childmask for the current node
			childMask = dag[offset]; 
			// Save the node offset to stack if we hit the top common level
			if(level==topBacktrackLevel) topBacktrackStackEntry = offset;
			// Save the node offset to stack if we are not past the bottom
			// common level. 
			if(level<=bottomBacktrackLevel) bottomBacktrackStackEntry = offset;
			///////////////////////////////////////////////////////////////////
			// Fetch the visibility for the child node our path resides in
			// (see Section 6.2 and Figure 5)
			///////////////////////////////////////////////////////////////////
			int lvlbit = 1<<(LEVELS-level-2);
			int childIndex = ((leafPath.x & lvlbit) ? 2 : 0) + 
							 ((leafPath.y & lvlbit) ? 4 : 0) +
							 ((leafPath.z & lvlbit) ? 8 : 0);
			childVisibility = NodeVisibility(0x3 & (childMask >> (childIndex)));
			///////////////////////////////////////////////////////////////////
			// If the child is not entirely shadowed or visible, traverse 
			// further down. Count the number of preceeding, existing children
			// to know where the pointer to the child node lies in relation
			// to the current. 
			///////////////////////////////////////////////////////////////////
			if (childVisibility == PARTIAL){  
				unsigned int maskedChildMask = childMask & (0xAAAA >> (16-childIndex));
				unsigned int childOffset = __popc(maskedChildMask); 
				lastOffset = dag[offset + 1 + childOffset]; 
			}
			else break;
			level += 1; 
		}

		///////////////////////////////////////////////////////////////////////
		// If we need to traverse further, take one more step. This last level
		// before the leaf nodes is stored differently, as explained in 
		// Section 5.
		///////////////////////////////////////////////////////////////////////
		if(childVisibility == PARTIAL)
		{
			unsigned int childMask = dag[lastOffset]; 
			int childIndex = (leafPath.z & 0x7);
			childIndex *=2;
			childVisibility = NodeVisibility(0x3 & (childMask >> (childIndex)));
			if (childVisibility == PARTIAL){ 
				unsigned int maskedChildMask = childMask & (0xAAAA >> (16-childIndex));
				unsigned int childOffset = __popc(maskedChildMask); 
				lastOffset = dag[lastOffset + 1 + childOffset];
			}
		}

		///////////////////////////////////////////////////////////////////////
		// Time to evaluate the number of samples that are both inside the 
		// filter region and in shadow. 
		//
		// First find the (2D) paths to the lower left, and upper right corners 
		// of the filter. 
		///////////////////////////////////////////////////////////////////////
		const int2 FilterLLPath = make_int2(positionPath) - make_int2(filterWidthAndHeight/2);
		const int2 FilterURPath = FilterLLPath + make_int2(filterWidthAndHeight-1); 
		///////////////////////////////////////////////////////////////////////
		// Then the lower left and upper right corners of the current leaf node
		///////////////////////////////////////////////////////////////////////
		int2 currentLeafLowerLeft = make_int2(x&(~0x7),y&(~0x7));
		int2 currentLeafUpperRight = make_int2(x|0x7, y|0x7);
		///////////////////////////////////////////////////////////////////////
		// The intersection of these is the area we are interested in.
		///////////////////////////////////////////////////////////////////////
		int2 lowerLeft = max(FilterLLPath, currentLeafLowerLeft);
		int2 upperRight = min(FilterURPath, currentLeafUpperRight);		
		///////////////////////////////////////////////////////////////////////
		// Calculate the total number of samples in the intersection
		///////////////////////////////////////////////////////////////////////
		int2 intersectionSize = make_int2(max(0, ((upperRight.x + 1) - lowerLeft.x)), 
										  max(0, ((upperRight.y+1) - lowerLeft.y)));
		int texelsInIntersection = intersectionSize.x * intersectionSize.y; 
		int visibleSamplesInIntersection = 0; 
		///////////////////////////////////////////////////////////////////////
		// Now establish how many of these samples are in shadow
		///////////////////////////////////////////////////////////////////////
		if(texelsInIntersection > 0)
		{
			if(childVisibility == SHADOW) { visibleSamplesInIntersection = 0; } 
			if(childVisibility == VISIBLE) { visibleSamplesInIntersection = texelsInIntersection;}
			if(childVisibility == PARTIAL) { 
					///////////////////////////////////////////////////////////
					// Coordinates within the leaf
					///////////////////////////////////////////////////////////
					int2 LL = lowerLeft & make_int2(0x7);   
					int2 UR = upperRight & make_int2(0x7); 
					///////////////////////////////////////////////////////////
					// Create a bitmask with set bits for all samples that are
					// in the intersection
					///////////////////////////////////////////////////////////
					unsigned char rows = (0xFF>>LL.x) & (0xFF<<(7-UR.x));
					uint2 cols = make_uint2(0x01010101, 0x01010101) & 
						make_uint2(0xFFFFFFFF << (LL.y*8),  0xFFFFFFFF << ((((LL.y-4)<0)?0:(LL.y-4))*8)) &
						make_uint2(0xFFFFFFFF >> ((((3-UR.y)>0)?(3-UR.y):0)*8), 0xFFFFFFFF >> ((7-UR.y)*8));
					uint2 intersectionMask = make_uint2(rows*cols.x, rows*cols.y);
					///////////////////////////////////////////////////////////
					// Fetch the 64bit leafmask and AND it with the mask for 
					// the intersection. 
					///////////////////////////////////////////////////////////
					uint2 leafMask = make_uint2(dag[lastOffset + 0], dag[lastOffset + 1]);
					uint2 shadowBits = leafMask & intersectionMask; 
					///////////////////////////////////////////////////////////
					// Count the number of set bits in the result
					///////////////////////////////////////////////////////////
					int setBits = (__popc(shadowBits.x) + __popc(shadowBits.y));
					visibleSamplesInIntersection = texelsInIntersection - setBits;
			}
		}
		///////////////////////////////////////////////////////////////////////
		// Accumulate visible samples within filter from this leaf 
		///////////////////////////////////////////////////////////////////////
		finalVisibility += visibleSamplesInIntersection;
		///////////////////////////////////////////////////////////////////////
		// Update path with next leaf node in traversal order
		///////////////////////////////////////////////////////////////////////
		lastPath = leafPath;
		x = LLPath.x + (travOrder & 0x6)*4;
		y = LLPath.y + (travOrder & 0x18);
		///////////////////////////////////////////////////////////////////////
		// If the traversal order is empty, we are done, otherwise shift it 
		// down. 
		///////////////////////////////////////////////////////////////////////
		thereAreLeafnodesLeft = (travOrder != 0);
		travOrder >>= 4;
	}
	///////////////////////////////////////////////////////////////////////////
	// Having visited all leaf nodes that are touched by the filter, return 
	// the proportion of filter-samples that were visible
	///////////////////////////////////////////////////////////////////////////
	return finalVisibility * (1.0f/float(filterWidthAndHeight * filterWidthAndHeight));
}

///////////////////////////////////////////////////////////////////////////////
// renderShadows_kernel - Fetch the light's NDC coordinate for one view sample
//                        evaluate visibility from the DAG at that point and 
//                        write result to texture. 
///////////////////////////////////////////////////////////////////////////////
template<int LEVELS>
__global__ void renderShadows_kernel(const unsigned int * __restrict__ dag, 
									 const unsigned int * __restrict__ topGrid,
									 unsigned int width, 
									 unsigned int height,
									 int filterSize,
									 matrix cuda_light_mvp,
									 float constant_bias)
{
	// Calculate pixel coordinates for this thread. Discard off-screen pixels. 
	int2 coord = make_int2(blockIdx.x * blockDim.x + threadIdx.x, 
						   blockIdx.y * blockDim.y + threadIdx.y);

	if(coord.x >= width || coord.y >= height) return; 
	// Get light-NDC coordinates for the view-sample. Discard if cleared pixel.
	float4 position = tex2D(positionTexture, float(coord.x), float(coord.y)); 
	float3 normal = make_float3(tex2D(normalTexture, float(coord.x), float(coord.y))); 
	if(position.w == 0.0f) { return; }
	// Using a constant bias from the normal here, which might not work
	// as well as the dxdy:ed thing we use in the paper code.
	float3 biasVec = normal * constant_bias; 
	position.x += biasVec.x; 
	position.y += biasVec.y; 
	position.z += biasVec.z; 
	float4 ndc = matrixmul(make_float3(position), cuda_light_mvp); 
	ndc *= (1.0f/ndc.w); 
	// Perform visibility lookup
	float visibility = shadowLookup<LEVELS>(make_float3(ndc),	  // Light NDC coords of viewsample, 
											dag,				  // Device pointer to DAG
											topGrid,			  // Device pointer to top grid
											filterSize			  // Actual filter size (can be dynamic)
											);
	// Write result to texture for shading
	surf2Dwrite(visibility, shadowSurface, coord.x*sizeof(float), coord.y);
}


///////////////////////////////////////////////////////////////////////////////
// renderShadows - Main entry function. Starts one thread per viewsample. Each 
//                 thread evaluates visibility for that view-sample and writes
//                 the result to a texture, which is then used in shading. 
//                 
//                 The DAG is expected to be loaded, the top-grid built and
//                 the textures (positionSurface and shadowSurface) registered
//                 before running this method. 
//
//				   Returns the meassured time 
///////////////////////////////////////////////////////////////////////////////
void CompressedShadow::renderShadows(const chag::float4x4 &light_mvp)
{
	matrix cuda_light_mvp; 
	memcpy(&cuda_light_mvp, &light_mvp.c1.x, sizeof(float)*16);

	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<13>, cudaFuncCachePreferL1); 
	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<14>, cudaFuncCachePreferL1); 
	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<15>, cudaFuncCachePreferL1); 
	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<16>, cudaFuncCachePreferL1); 
	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<17>, cudaFuncCachePreferL1); 
	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<18>, cudaFuncCachePreferL1); 
	_CHECKED_ cudaFuncSetCacheConfig(renderShadows_kernel<19>, cudaFuncCachePreferL1); 

	cudaEvent_t beginEvent;
	cudaEvent_t endEvent;
	cudaEventCreate( &beginEvent );
	cudaEventCreate( &endEvent );

	// Map OpenGL textures for CUDA
	mapTextures();
	{	
		cudaEventRecord( beginEvent, 0 );

		// Set up one thread per view-sample
		dim3 blockDim = dim3(8, 16);
		dim3 gridDim = dim3((m_width + blockDim.x - 1) / blockDim.x, (m_height + blockDim.y - 1) / blockDim.y);
		// Launch kernel
		switch(m_levels) {
		///////////////////////////////////////////////////////////////////////
		// Kernels are precompiled for different DAG depths in order to avoid
		// an unseemly number of computations that can be static. A cleaner
		// approach would be JIT compilation. 
		///////////////////////////////////////////////////////////////////////
		case 13: renderShadows_kernel<13><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		case 14: renderShadows_kernel<14><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		case 15: renderShadows_kernel<15><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		case 16: renderShadows_kernel<16><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		case 17: renderShadows_kernel<17><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		case 18: renderShadows_kernel<18><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		case 19: renderShadows_kernel<19><<< gridDim, blockDim >>>(m_d_data, m_d_topGrid, m_width, m_height, m_filter_size, cuda_light_mvp, m_bias); break;
		default: LOG_ERROR("Expecting 4k-256k sized DAGS only."); break;
		}
		cudaEventRecord( endEvent, 0 );
	}
	// Unmap OpenGL textures
	unmapTextures(); 

	cudaEventSynchronize( endEvent );
	cudaEventElapsedTime( &m_meassured_time, beginEvent, endEvent );
}