#pragma once
#include <cuda.h>
#include <cuda_runtime.h>
#include <GL/glew.h>
#include <iostream>
#include <linmath/linmath.h>
#include <vector>
#include <scene/Scene.h>

#define _CHECKED_ CUDA_CHECKED_CALL
float3 inline make_float3(chag::float3 aVec) { return make_float3(aVec.x, aVec.y, aVec.z); }
#define TOPGRID_LEVELS 8

///////////////////////////////////////////////////////////////////////////////
// The CompressedShadow class contains all the code required for loading and 
// rendering with a CPVS.  It inherits the ShadowMapShadowGenerator only in 
// order to be used with the UI. 
///////////////////////////////////////////////////////////////////////////////
class CompressedShadow : public Scene::ShadowMapShadowGenerator
{
public: 
	// Number of levels in the DAG
	int m_levels; 
	// Device pointer to the shadow DAG 
	unsigned int *m_d_data; 
	// Device pointer to the top grid
	unsigned int *m_d_topGrid;
	// A simple constant bias in the normals direction
	float m_bias; 
	// The wanted filter size
	int m_filter_size; 
	// The file that the DAG was loaded from
	std::string m_file_name;
	// Size of the read data
	int m_data_size;
	// The actual DAG data read from file
	unsigned int *m_data;
	// A couple of levels of the DAG are kept in a grid for improved performance
	unsigned int *m_topGrid;
	// Width and height of rendertarget
	static unsigned int m_width, m_height;
	// CUDA handles for the gbuffer and final shadow buffer
	static cudaGraphicsResource *m_shadowResource;
	static cudaArray			 *m_shadowArray;
	static cudaGraphicsResource *m_positionResource;
	static cudaArray			 *m_positionArray;
	static cudaGraphicsResource *m_normalResource;
	static cudaArray			 *m_normalArray;
	// Time taken for lookups in last frame
	float m_meassured_time; 

	CompressedShadow(); 
	void renderShadows(const chag::float4x4 &light_mvp);
	void load(const std::string &filename); 
	void buildTopGrid(); 
	void uploadData(); 
	void debugRender(); 
	static void registerTextures(GLuint shadowTexture, GLuint positionTexture, GLuint normalTexture);
	static void unregisterTextures();
	static void resize(unsigned int w, unsigned int h) { m_width = w; m_height = h; };
	void mapTextures(); 
	void unmapTextures(); 

	///////////////////////////////////////////////////////////////////////////
	// Below is required only by Scene::ShadowGenerator (UI)
	///////////////////////////////////////////////////////////////////////////
	virtual Scene::Item *clone() { LOG_INFO("CompressedShadow::clone() not yet implemented."); return NULL; }; 
	virtual void save(std::ofstream &os) { 
		os << "compressedshadowgenerator " << m_name << " " << m_file_name << " " << m_light->m_name << " " << m_fov << " " << m_aspect_ratio << " " 
				<< m_near << " " << m_far << " " << m_bias << " " << m_filter_size << std::endl; 
	}; 
	virtual void update(const std::vector<Scene::Model *> &models, GLSLProgramObject *program) {};
};
