#include "Quaternion.h"
#include "float3x3.h"
#include <cmath>

namespace chag
{



Quaternion make_quaternion_axis_angle(const float3 &_axis, float angle)
{
  float3 axis = normalize(_axis);

  float halfAngle = 0.5f * angle;
  float sinHalfAngle = sinf(halfAngle);
  return Quaternion(axis * sinHalfAngle, cosf(halfAngle));
}



float n(const Quaternion& q) 
{
  return std::sqrt(q.v.x * q.v.x + q.v.y * q.v.y + q.v.z * q.v.z + q.w * q.w);
};



const Quaternion inv(const Quaternion& q)
{
  return q.conj() * (1.0f / n(q));
}

const float4x4 makematrix(const Quaternion& q)
{
  // This should be simplified for unit quaternions
  float4x4 m;
  float s = 2.0f / n(q);
  m.c1 = make_vector(1.0f - s*(q.v.y*q.v.y + q.v.z*q.v.z), 
		  s*(q.v.x*q.v.y + q.w*q.v.z), 
		  s*(q.v.x*q.v.z - q.w*q.v.y), 
		  0.0f);
  m.c2 = make_vector(s*(q.v.x*q.v.y - q.w*q.v.z), 
		  1.0f - s*(q.v.x*q.v.x + q.v.z*q.v.z), 
		  s*(q.v.y*q.v.z + q.w*q.v.x), 
		  0.0f);
  m.c3 = make_vector(s*(q.v.x*q.v.z + q.w*q.v.y), 
		  s*(q.v.y*q.v.z - q.w*q.v.x), 
		  1.0f - s*(q.v.x*q.v.x + q.v.y*q.v.y), 
		  0.0f);
  m.c4 = make_vector(0.0f, 0.0f, 0.0f, 1.0f);
  return m;
}



const float3x3 toMatrix3x3(const Quaternion& q)
{
  // This should be simplified for unit quaternions
  float3x3 m;
  float s = 2.0f / n(q);
  m.c1 = make_vector(1.0f - s*(q.v.y*q.v.y + q.v.z*q.v.z), 
		  s*(q.v.x*q.v.y + q.w*q.v.z), 
		  s*(q.v.x*q.v.z - q.w*q.v.y));
  m.c2 = make_vector(s*(q.v.x*q.v.y - q.w*q.v.z), 
		  1.0f - s*(q.v.x*q.v.x + q.v.z*q.v.z), 
		  s*(q.v.y*q.v.z + q.w*q.v.x));
  m.c3 = make_vector(s*(q.v.x*q.v.z + q.w*q.v.y), 
		  s*(q.v.y*q.v.z - q.w*q.v.x), 
		  1.0f - s*(q.v.x*q.v.x + q.v.y*q.v.y));
  return m;
}


const float phi(Quaternion q, Quaternion r)
{
	float qdotr = q.v.x*r.v.x + q.v.y*r.v.y + q.v.z*r.v.z + q.w*r.w;
	return std::acos(qdotr);
}

const Quaternion slerp(Quaternion q, Quaternion r, float t)
{
	double cosHalfTheta = q.w * r.w + q.v.x * r.v.x + q.v.y * r.v.y + q.v.z * r.v.z;

	// While q and -q represent the same orientation, the slerp between em 
	// differs. Common solution is apparently to theck q*r > 0 and negate. 
	if(cosHalfTheta < 0.0)
	{
		r.w = -r.w; 
		r.v = -r.v;
		cosHalfTheta = q.w * r.w + q.v.x * r.v.x + q.v.y * r.v.y + q.v.z * r.v.z;
	}

	// if q=r or q=-r then theta = 0 and we can return q
	if (std::abs(cosHalfTheta) >= 0.999)
	{
		// lerp
		Quaternion ret; 
		ret.v = lerp(q.v, r.v, t);
		ret.w = lerp(q.w, r.w, t);
		
		float4 a = normalize(make_vector(ret.v.x, ret.v.y, ret.v.z, ret.w)); 
		ret.v = make_vector(a.x,a.y,a.z);
		ret.w = a.w; 

		return ret; 
	}

	// Calculate temporary values.
	double halfTheta = std::acos(cosHalfTheta);

	double sinHalfTheta = sqrt(1.0 - cosHalfTheta*cosHalfTheta);
	// if theta = 180 degrees then result is not fully defined
	// we could rotate around any axis normal to qa or qb
	// Note: The below code did NOT test for this condition!
	//if (std::abs(sinHalfTheta) < 0.001)
	//{
	//            Quaternion qn = (q * 0.5) + (r * 0.5);
	//            float mag = sqrt(dot(qn.v, qn.v) + square(qn.w));
	//            qn.v /= mag;
	//            qn.w /= mag;
	//            return qn;
	//}

	double ratioA = std::sin((1 - t) * halfTheta) / sinHalfTheta;
	double ratioB = std::sin(t * halfTheta) / sinHalfTheta; 
	//calculate Quaternion.
	return q * float(ratioA) + r * float(ratioB);

}


float ReciprocalSqrt( float x ) {
	long i;
	float y, r;
	y = x * 0.5f;
	i = *(long *)( &x );
	i = 0x5f3759df - ( i >> 1 );
	r = *(float *)( &i );
	r = r * ( 1.5f - r * r * y );
	return r;
}

Quaternion make_quaternion(const float3x3 &mat3x3)
//	ConvertJointMatsToJointQuats( JointQuat *jointQuats, const JointMat *jointMats, const int numJoints ) 
{
	float q[4];
	float4x4 mat = make_matrix(mat3x3, make_vector(0.0f, 0.0f, 0.0f)); 
	const float *m = &mat.c1.x; 
	if ( m[0 * 4 + 0] + m[1 * 4 + 1] + m[2 * 4 + 2] > 0.0f ) {
		float t = + m[0 * 4 + 0] + m[1 * 4 + 1] + m[2 * 4 + 2] + 1.0f;
		float s = ReciprocalSqrt( t ) * 0.5f;
		q[3] = s * t;
		q[2] = ( m[0 * 4 + 1] - m[1 * 4 + 0] ) * s;
		q[1] = ( m[2 * 4 + 0] - m[0 * 4 + 2] ) * s;
		q[0] = ( m[1 * 4 + 2] - m[2 * 4 + 1] ) * s;
	} else if ( m[0 * 4 + 0] > m[1 * 4 + 1] && m[0 * 4 + 0] > m[2 * 4 + 2] ) {
		float t = + m[0 * 4 + 0] - m[1 * 4 + 1] - m[2 * 4 + 2] + 1.0f;
		float s = ReciprocalSqrt( t ) * 0.5f;
		q[0] = s * t;
		q[1] = ( m[0 * 4 + 1] + m[1 * 4 + 0] ) * s;
		q[2] = ( m[2 * 4 + 0] + m[0 * 4 + 2] ) * s;
		q[3] = ( m[1 * 4 + 2] - m[2 * 4 + 1] ) * s;
	} else if ( m[1 * 4 + 1] > m[2 * 4 + 2] ) {
		float t = - m[0 * 4 + 0] + m[1 * 4 + 1] - m[2 * 4 + 2] + 1.0f;
		float s = ReciprocalSqrt( t ) * 0.5f;
		q[1] = s * t;
		q[0] = ( m[0 * 4 + 1] + m[1 * 4 + 0] ) * s;
		q[3] = ( m[2 * 4 + 0] - m[0 * 4 + 2] ) * s;
		q[2] = ( m[1 * 4 + 2] + m[2 * 4 + 1] ) * s;
	} else {
		float t = - m[0 * 4 + 0] - m[1 * 4 + 1] + m[2 * 4 + 2] + 1.0f;
		float s = ReciprocalSqrt( t ) * 0.5f;
		q[2] = s * t;
		q[3] = ( m[0 * 4 + 1] - m[1 * 4 + 0] ) * s;
		q[0] = ( m[2 * 4 + 0] + m[0 * 4 + 2] ) * s;
		q[1] = ( m[1 * 4 + 2] + m[2 * 4 + 1] ) * s;
	}
	Quaternion r; 
	r.v.x = q[0]; 
	r.v.y = q[1]; 
	r.v.z = q[2]; 
	r.w = q[3]; 
	return r; 
}





} // namespace chag
