#include "Scene.h"
#include <utils/Log.h>
#include <sstream>

using namespace chag; 
using namespace std; 

Scene::Scene(void)
{
	m_controllableViewCounter = 0; 
	m_renderableViewCounter = 0; 
	m_sceneUnitMultiplier = 1.0f; 
	{
		PerspectiveCamera *camera = new PerspectiveCamera; 
		camera->m_name = "DefaultCamera";
		camera->m_transform.lookAt(make_vector(0.0f, 0.0f, -10.0f), make_vector(0.0f, 0.0f, 0.0f), make_vector(0.0f, 1.0f, 0.0f));
		camera->m_near = 0.1f; 
		camera->m_far = 1000.0f; 
		camera->m_fov = 60.0f; 
		camera->m_aspect_ratio = 1.0f; 
		m_cameras.push_back(camera); 
	}
	{
		DiscDiffuseLight *light = new DiscDiffuseLight; 
		light->m_name = "DefaultLight"; 
		light->m_transform.lookAt(make_vector(0.0f, 10.0f, 0.0f), make_vector(0.0f, -1.0f, 0.0f));
		light->m_color = make_vector(1.0f, 1.0f, 1.0f);
		light->m_radiance = 1000.0f; 
		light->m_radius = 1.0f; 
		m_lights.push_back(light); 
	}

	m_programManager = new GLSLProgramManager("../Scene/Scene.glsl"); 
	m_shadowMapProgram = m_programManager->getProgramObject("shadowMapVertex", "shadowMapFragment"); 
}

Scene::~Scene(void)
{
}

bool Scene::save(string filename) 
{
	if(filename == "") filename = m_filename; 
	ofstream os(filename.c_str(), ios::trunc); 
	if(!os.is_open()) {
		LOG_ERROR("Failed to save " << filename); 
		return false; 
	}
	os << "sceneUnitMultiplier " << m_sceneUnitMultiplier << endl; 
	for(auto camera : m_cameras) {
		os << *camera << endl; 
	}
	for(auto light : m_lights) {
		os << *light << endl; 
	}
	for(auto model : m_models) {
		os << *model << endl;
		for(auto mat : model->m_materials) {
			os << "material " << model->m_name << " " << *mat.second << endl; //<< mat.second->m_name << " " << mat.second->m_diffuse_reflectance << endl; 
		}		
	}
	for(auto sg : m_shadowGenerators) {
		sg->save(os); 
	}
	os.close();
	LOG_INFO("Saved " << filename); 
	return true;
}

bool Scene::load(string filename, SceneExtender  *scene_extender)
{
	ifstream is(filename.c_str()); 
	if(!is.is_open()) {
		LOG_ERROR("Failed to load " << filename);
		return false; 
	}
	for(auto camera : m_cameras) delete camera; 
	for(auto light : m_lights) delete light; 
	for(auto model : m_models) delete model; 
	m_cameras.clear(); 
	m_lights.clear(); 
	m_models.clear(); 
	while(!is.eof()) {
		char buffer[4096]; 
		is.getline(buffer, 4096); 
		string line(buffer); 
		stringstream ss(line); 
		string type; 
		ss >> type; 
		if(type == "perspective_camera") {
			Scene::PerspectiveCamera *camera = new Scene::PerspectiveCamera; 
			ss >> *camera; 
			m_cameras.push_back(camera); 
		}
		else if(type == "disc_diffuse_light") {
			Scene::DiscDiffuseLight *light = new Scene::DiscDiffuseLight;
			ss >> *light; 
			m_lights.push_back(light); 
		}
		else if(type == "model") {
			Scene::Model *model = new Scene::Model; 
			ss >> *model; 
			// Allright, the rest of this could be prettier.
			// Why isn't loadOBJ() a member function of Model?
			Scene::Model *objModel = loadOBJModel(model->m_filename); 
			objModel->m_name = model->m_name; 
			objModel->m_transform = model->m_transform; 
			delete model; 
			m_models.push_back(objModel); 
		}
		else if(type == "material") {
			string modelName; 
			ss >> modelName; 
			string materialName; 
			ss >> materialName; 
			Material *mat = getModel(modelName)->m_materials[materialName]; 
			if(!mat) {
				mat = new Material; 
				getModel(modelName)->m_materials[materialName] = mat; 
			}
			mat->m_name = materialName; 
			ss >> *mat; 
		}
		else if(type == "sceneUnitMultiplier") {
			ss >> m_sceneUnitMultiplier; 
		}
		else if(type == "shadowmapshadowgenerator") {
			ShadowMapShadowGenerator *new_sg = new ShadowMapShadowGenerator(); 
			ss >> new_sg->m_name; 
			string lightName; 
			ss >> lightName; 
			new_sg->m_light = NULL; 
			for(auto l : m_lights) { if(l->m_name == lightName) new_sg->m_light = l; }
			if(new_sg->m_light == NULL) LOG_ERROR("Light " << lightName << "not found for shadowgenerator"); 
			ss  >> new_sg->m_fov >> new_sg->m_aspect_ratio >> new_sg->m_near >> new_sg->m_far 
				>> new_sg->m_polygon_offset_factor >> new_sg->m_polygon_offset_units >> new_sg->m_resolution.x >> new_sg->m_resolution.y; 
			new_sg->resize(new_sg->m_resolution.x, new_sg->m_resolution.y); 
			m_shadowGenerators.push_back(new_sg); 
		}
		else {
			if(scene_extender) {
				scene_extender->handleUnknownType(type, ss, this); 
			}
			else {
				LOG_ERROR("Unexpected type found in scene '" << filename << "': '" << type << "'");
			}
		}
	}
	m_filename = filename; 
	return true; 
}

bool Scene::importOBJ(string filename) 
{
	Scene::Model *model = loadOBJModel(filename, true); 
	stringstream ss; 
	ss << filename; 
	model->m_name = ss.str(); 
	model->m_filename = filename; 
	m_models.push_back(model); 
	return true; 
}


void Scene::Model::render(GLSLProgramObject *programObject)
{
	programObject->setUniform("modelMatrix", m_transform.get_MV_inv()); 
	glBindVertexArray(m_vaob);
	for(auto material : m_materials) {
		float diffuse_weight = material.second->m_diffuse_weight / (material.second->m_diffuse_weight + material.second->m_glossy_weight); 
		float glossy_weight = material.second->m_glossy_weight / (material.second->m_diffuse_weight + material.second->m_glossy_weight); 
		programObject->setUniform("reflective_coating_enable",		material.second->m_reflective_coating_enable); 
		programObject->setUniform("reflective_coating_reflectance", material.second->m_reflective_coating_reflectance); 
		programObject->setUniform("reflective_coating_roughness",	material.second->m_reflective_coating_roughness); 
		programObject->setUniform("reflective_coating_0_degrees",	material.second->m_reflective_coating_0_degrees); 
		programObject->setUniform("reflective_coating_90_degrees",	material.second->m_reflective_coating_90_degrees); 
		programObject->setUniform("diffuse_weight",					diffuse_weight); 
		programObject->setUniform("diffuse_reflectance",			material.second->m_diffuse_reflectance); 
		programObject->setUniform("glossy_weight",					glossy_weight); 
		programObject->setUniform("glossy_reflectance",				material.second->m_glossy_reflectance); 
		programObject->setUniform("glossy_roughness",				material.second->m_glossy_roughness); 
		if(m_drawCallOffsets.count(material.second) > 0) {
			glMultiDrawArrays(GL_TRIANGLES, &m_drawCallOffsets[material.second][0], &m_drawCallCounts[material.second][0], int(m_drawCallCounts[material.second].size())); 
		}
	}
}

Aabb Scene::Model::getAabb()
{	
	LOG_WARNING("Scene::Model::getAabb not implemented"); 
	return make_inverse_extreme_aabb(); 
}

void Scene::submitModels(GLSLProgramObject *programObject)
{
	for(int i=0; i<m_models.size(); i++) {
		m_models[i]->render(programObject); 
	}
}

void Scene::submitLightGeometry(GLSLProgramObject *programObject)
{
	for(auto light : m_lights) {
		DiscDiffuseLight *ddl = dynamic_cast<DiscDiffuseLight *>(light); 
		if(ddl) {
			glBegin(GL_TRIANGLES); 
			const int SEGMENTS = 16; 
			for(int i=0; i<SEGMENTS; i++) {
				glColor3f(ddl->m_color.x, ddl->m_color.y, ddl->m_color.z); 
				float alpha = (float(i)/float(SEGMENTS)) * 2.0f * 3.141592f; 
				float dalpha = 1.0f/float(SEGMENTS) * 2.0f * 3.141592f; 
				float3 &position = ddl->m_transform.pos; 
				float3 &direction = ddl->m_transform.R.c3; 
				float3 &tangent = ddl->m_transform.R.c1; 
				float3 &bitangent = ddl->m_transform.R.c2; 
				glVertex3fv(&position.x);
				float3 p2 = position + ddl->m_radius * normalize(sin(alpha) * tangent + cos(alpha) * bitangent); 
				glVertex3fv(&p2.x); 
				float3 p3 = position + ddl->m_radius * normalize(sin(alpha + dalpha) * tangent + cos(alpha + dalpha) * bitangent); 
				glVertex3fv(&p3.x); 
			}
			glEnd(); 
		}
	}
}

void Scene::submitLights(GLSLProgramObject *programObject, orientation &transform)
{
	int number_of_lights = int(m_lights.size());
	vector<float3> light_position; 
	vector<float3> light_direction; 
	vector<float>  light_radiance; 
	vector<float3> light_color; 
	vector<float>  light_radius; 
	programObject->setUniform("lights_number_of_lights", number_of_lights); 
	float4x4 MV = transform.get_MV(); 
	float3x3 MVInvTransp = transpose(make_matrix3x3(transform.get_MV_inv()));
	for(int i=0; i<m_lights.size(); i++) {
		DiscDiffuseLight *ddl = dynamic_cast<DiscDiffuseLight *>(m_lights[i]); 
		if(ddl) {
			light_position.push_back(make_vector3(MV * make_vector4(ddl->m_transform.pos, 1.0f))); 
			light_direction.push_back(-(MVInvTransp * ddl->m_transform.R.c3)); 
			light_radiance.push_back(ddl->m_radiance); 
			light_color.push_back(ddl->m_color); 
			light_radius.push_back(ddl->m_radius); 
		}
		else {
			LOG_ERROR("Non DiscDiffuseLight not implemented");
		}
	}
	programObject->setUniform("light_position", (float3 *)&light_position[0].x, number_of_lights); 
	programObject->setUniform("light_direction", (float3 *)&light_direction[0].x, number_of_lights); 
	programObject->setUniform("light_radiance", &light_radiance[0], number_of_lights); 
	programObject->setUniform("light_color", (float3 *)&light_color[0].x, number_of_lights); 
	programObject->setUniform("light_radius", &light_radius[0], number_of_lights); 

	float3x3 MVtranspose = make_matrix3x3(transpose(transform.get_MV())); 
	programObject->setUniform("modelViewTranspose", MVtranspose); 

	if(m_environmentMap.m_active) {
		programObject->setUniform("environment_map_active", true); 
		programObject->bindTexture("environment_map_texture", m_environmentMap.m_texture_id, GL_TEXTURE_CUBE_MAP, 0); 
	}
}

Aabb Scene::getAabb()
{
	Aabb r = make_inverse_extreme_aabb(); 
	for(int i=0; i<m_models.size(); i++) {
		r = combine(r, m_models[i]->getAabb()); 
	}
	return r; 
}

Scene::Item * Scene::getCurrentControllableItem() {
	if(m_controllableViewCounter < m_cameras.size()) return m_cameras[m_controllableViewCounter]; 
	else if(m_controllableViewCounter < (m_cameras.size() + m_lights.size())) return m_lights[m_controllableViewCounter - m_cameras.size()]; 
	else return m_models[m_controllableViewCounter -(m_cameras.size() + m_lights.size())]; 
}

void Scene::nextControllableItem() {
	m_controllableViewCounter = (m_controllableViewCounter + 1) % (m_cameras.size() + m_lights.size() + m_models.size()); 
}; 

void Scene::prevControllableItem() {
	m_controllableViewCounter = (m_controllableViewCounter - 1) % (m_cameras.size() + m_lights.size() + m_models.size()); 
}
Scene::Item * Scene::getCurrentRenderableItem() {
	if(m_renderableViewCounter < m_cameras.size()) return m_cameras[m_renderableViewCounter]; 
	else if(m_renderableViewCounter < (m_cameras.size() + m_lights.size())) return m_lights[m_renderableViewCounter - m_cameras.size()]; 
	else if(m_renderableViewCounter < (m_cameras.size() + m_lights.size() + m_models.size())) return m_models[m_renderableViewCounter - (m_cameras.size() + m_lights.size())];  
	else return m_shadowGenerators[m_renderableViewCounter - (m_cameras.size() + m_lights.size() + m_models.size())];  
}
void Scene::nextRenderableItem() {
	m_renderableViewCounter = (m_renderableViewCounter + 1) % (m_cameras.size() + m_lights.size() + m_models.size() + m_shadowGenerators.size()); 
}
void Scene::prevRenderableItem() {
	m_renderableViewCounter = (m_renderableViewCounter - 1) % (m_cameras.size() + m_lights.size() + m_models.size() + m_shadowGenerators.size()); 
}

void Scene::setCurrentControllableItem(const std::string &name) 
{
	for(unsigned int i=0; i<m_cameras.size(); i++) if(m_cameras[i]->m_name == name) m_controllableViewCounter = i; 
	for(unsigned int i=0; i<m_lights.size(); i++) if(m_lights[i]->m_name == name) m_controllableViewCounter = int(m_cameras.size()) + i; 
	for(unsigned int i=0; i<m_models.size(); i++) if(m_models[i]->m_name == name) m_controllableViewCounter = int(m_cameras.size()) + int(m_lights.size()) + i; 
}

void Scene::setCurrentRenderableItem(const std::string &name) 
{
	for(int i=0; i<m_cameras.size(); i++) if(m_cameras[i]->m_name == name) m_renderableViewCounter = i; 
	for(int i=0; i<m_lights.size(); i++) if(m_lights[i]->m_name == name) m_renderableViewCounter = int(m_cameras.size()) + i; 
	for(int i=0; i<m_models.size(); i++) if(m_models[i]->m_name == name) m_renderableViewCounter = int(m_cameras.size()) + int(m_lights.size()) + i; 
	for(int i=0; i<m_shadowGenerators.size(); i++) if(m_shadowGenerators[i]->m_name == name) m_renderableViewCounter = int(m_cameras.size()) + int(m_lights.size() + int(m_models.size())) + i; 
}

void Scene::EnvironmentMap::load(unsigned char *posx, unsigned char *posy, unsigned char *posz,
								 unsigned char *negx, unsigned char *negy, unsigned char *negz,
								 unsigned int width, unsigned int height)
{
	m_width = width; 
	m_height = height; 
	vector<unsigned char *> image; 
	image.push_back(posx); 	image.push_back(posy); 	image.push_back(posz); 
	image.push_back(negx); 	image.push_back(negy); 	image.push_back(negz); 
	if(m_faces[0] != NULL) glDeleteTextures(1, &m_texture_id); 
	for(int i=0; i<6; i++) { 
		if(m_faces[i] != NULL) delete [] m_faces[i]; 
		m_faces[i] = new float3[m_width * m_height];
		for(unsigned int j=0; j<m_width*m_height; j++) {
			m_faces[i][j] = make_vector(image[i][j*3+0]/float(255.0f), image[i][j*3+1]/float(255.0f), image[i][j*3+2]/float(255.0f));
		}
	}
	glGenTextures(1, &m_texture_id);
	glBindTexture(GL_TEXTURE_CUBE_MAP, m_texture_id); 
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_RGB32F, m_width, m_height, 0, GL_RGB, GL_FLOAT, m_faces[0]); 
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, GL_RGB32F, m_width, m_height, 0, GL_RGB, GL_FLOAT, m_faces[1]); 
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, GL_RGB32F, m_width, m_height, 0, GL_RGB, GL_FLOAT, m_faces[2]); 
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, GL_RGB32F, m_width, m_height, 0, GL_RGB, GL_FLOAT, m_faces[3]); 
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, GL_RGB32F, m_width, m_height, 0, GL_RGB, GL_FLOAT, m_faces[4]); 
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, GL_RGB32F, m_width, m_height, 0, GL_RGB, GL_FLOAT, m_faces[5]); 
	m_active = true; 
}

float Scene::DiscDiffuseLight::pdf(const float3 &p, const float3 &lightSample)
{
	float distanceSquared = length(p-lightSample) * length(p-lightSample); 
	float3 wi = normalize(p-lightSample); 
	float cosAngle = dot(wi, -m_transform.R.c3); 
	if(cosAngle <= 0.0f) return 0.0f; 
	float lightPdf = distanceSquared / (abs(cosAngle) * area());
	if(isinf(lightPdf)) 
		lightPdf = 0.0f; 
	return lightPdf; 
}

float3 Scene::DiscDiffuseLight::sample(const float3 &p, float &lightPdf)
{
	// Sample a uniform position on disc
	float2 discSample; 
	concentricSampleDisk(&discSample.x,&discSample.y); 
	// Gah... Should precompute these.
	float3 tangent = normalize(perpendicular(-m_transform.R.c3)); 
	float3 bitangent = normalize(cross(tangent, -m_transform.R.c3));
	float3 lightSample = m_transform.pos + m_radius * (discSample.x * tangent + discSample.y * bitangent);
	lightPdf = pdf(p, lightSample); 
	return lightSample; 
}

float3 Scene::DiscDiffuseLight::Le(const float3 &p, const float3 &wo)
{
	// Return radiance emitted from p in direction wo
	//float3 normal = normalize(p - m_position);
	//return m_power * (1.0f/area()) * (1.0f/(4.0f * M_PI));  
	if(dot(wo,-m_transform.R.c3) < 0.0f) return make_vector(0.0f, 0.0f, 0.0f); 
	return m_radiance * m_color; 
}

bool Scene::DiscDiffuseLight::intersect(const RTCRay &r, float &t) 
{
	// Intersect with plane
	float4 plane = make_vector4(-m_transform.R.c3, -dot(m_transform.pos, -m_transform.R.c3)); 
	float t0 = -dot(plane, make_vector4(asFloat3(r.org), 1.0f)) / dot(make_vector3(plane), asFloat3(r.dir)); 
	if(t0 < r.tnear || t0 > r.tfar) return false; 
	// Point on plane
	float3 p = onRay(r, t0); 
	if(length(p-m_transform.pos) > m_radius) return false; 
	t = t0; 
	return true; 
}

void Scene::initializeRaytracer()
{
	auto checkEmbreeError = [] (string &str) {
		if(rtcGetError() != RTC_NO_ERROR) {
			LOG_ERROR("Embree initialization failed."); 
		}
	}; 
	rtcInit(NULL);
	checkEmbreeError(string("rtcInit")); 
	m_embreeScene = rtcNewScene(RTC_SCENE_STATIC,RTC_INTERSECT1);
	checkEmbreeError(string("rtcNewScene")); 
	for(auto model : m_models) {
		model->m_embreeGeometryID = rtcNewTriangleMesh(m_embreeScene, RTC_GEOMETRY_STATIC, model->m_positions.m_host_vector.size()/3, model->m_positions.m_host_vector.size(), 1);
		checkEmbreeError(string("rtcNewTriangleMesh")); 
		vector<float4> embreeVertices; 
		vector<int3> embreeTriangles; 
		for(auto vertex : model->m_positions.m_host_vector) embreeVertices.push_back(make_vector4(vertex, 1.0f)); 
		for(int i=0; i<int(model->m_positions.m_host_vector.size())/3; i++) {
			embreeTriangles.push_back(make_vector(i*3 + 0, i*3 + 1, i*3 + 2)); 
		}
		float4 *v = (float4 *) rtcMapBuffer(m_embreeScene, model->m_embreeGeometryID, RTC_VERTEX_BUFFER);
		memcpy(v, &embreeVertices[0].x, embreeVertices.size() * sizeof(float4)); 
		rtcUnmapBuffer(m_embreeScene, model->m_embreeGeometryID, RTC_VERTEX_BUFFER); 
		int3 *t = (int3 *) rtcMapBuffer(m_embreeScene, model->m_embreeGeometryID, RTC_INDEX_BUFFER); 
		memcpy(t, &embreeTriangles[0].x, embreeTriangles.size() * sizeof(int3)); 
		rtcUnmapBuffer(m_embreeScene, model->m_embreeGeometryID, RTC_INDEX_BUFFER); 
		checkEmbreeError(string("Upload data")); 
		m_embreeIDToModelMap[model->m_embreeGeometryID] = model; 
	}
	rtcCommit(m_embreeScene); 
	checkEmbreeError(string("Commit")); 
}

bool Scene::intersectP(RTCRay &r) {
	rtcOccluded(m_embreeScene, r); 
	return r.geomID != RTC_INVALID_GEOMETRY_ID; 
}

bool Scene::intersect(RTCRay &r, Intersection &isect)
{
	rtcIntersect(m_embreeScene, r); 
	if(r.geomID != RTC_INVALID_GEOMETRY_ID) {
		Model *model = m_embreeIDToModelMap[r.geomID]; 
		float3 v0 = model->m_positions.m_host_vector[r.primID*3 + 0]; 
		float3 v1 = model->m_positions.m_host_vector[r.primID*3 + 1]; 
		float3 v2 = model->m_positions.m_host_vector[r.primID*3 + 2]; 
		isect.m_geometryNormal = normalize(cross((v1-v0), (v2-v0))); 
		isect.m_material = model->m_triangleMaterial[r.primID]; 
		float w = 1.0f - (r.u + r.v);
		float3 n0 = model->m_normals.m_host_vector[r.primID*3 + 0]; 
		float3 n1 = model->m_normals.m_host_vector[r.primID*3 + 1]; 
		float3 n2 = model->m_normals.m_host_vector[r.primID*3 + 2]; 
		isect.m_normal = normalize(w * n0 + r.u * n1 + r.v * n2);
		float2 uv0 = model->m_uvs.m_host_vector[r.primID*3 + 0]; 
		float2 uv1 = model->m_uvs.m_host_vector[r.primID*3 + 1]; 
		float2 uv2 = model->m_uvs.m_host_vector[r.primID*3 + 2]; 
		isect.m_uv = w * uv0 + r.u * uv1 + r.v * uv2;
		isect.m_position = onRay(r, r.tfar); 
		return true; 
	}
	else return false; 
}

///////////////////////////////////////////////////////////////////////////////
// Raytracing material
///////////////////////////////////////////////////////////////////////////////

float Scene::Material::D(const float3 &m, const float3 &n, const float alpha) const
{
	if(dot(m,n) <= 0.0) return 0.0f; 
	float cosThetaM = dot(m,n); 
	float cosThetaMsq = cosThetaM * cosThetaM; 
	float tanThetaMsq = (1.0f - cosThetaMsq)/cosThetaMsq; 
	float alphaSq = alpha * alpha; 
	return (PCF(abs(dot(m,n)))/(M_PI*alphaSq*cosThetaMsq*cosThetaMsq))*exp(-tanThetaMsq/alphaSq);
}

float Scene::Material::G1(const float3 &v, const float3 &m, const float3 &n, const float alpha) const
{
	if(dot(m,n) <= 0.0) return 0.0f; 
	float cosThetaV = dot(v,n); 
	float cosThetaVsq = cosThetaV * cosThetaV; 
	float tanThetaVsq = (1.0f / cosThetaVsq) - 1.0f; 
	float tanThetaV = sqrt(tanThetaVsq); 
	float a = 1.0f / (alpha * tanThetaV);
	float b = PCF(dot(v,m)/dot(v,n));
	if(a >= 1.6f) return b * 1.0f;
	else {
		return b * (3.535f*a + 2.181f*a*a) / (1.0f + 2.276f*a + 2.577f*a*a); 
	}
}


float3 Scene::Material::f(const float3 &wi, const float3 &wo, const Intersection &isect) const
{
	const float3 &n = isect.facingNormal(wi); 
	float3 hr = normalize(sign(dot(wo, n)) * (wo + wi));
	if(!sameHemisphere(wi, wo, n)) return make_vector(0.0f, 0.0f, 0.0f); 
	float diffuse_weight = m_diffuse_weight / (m_diffuse_weight + m_glossy_weight); 
	float glossy_weight = m_glossy_weight / (m_diffuse_weight + m_glossy_weight); 
	float fresnel = 0.0f; 
	float3 reflective_coating = make_vector(0.0f, 0.0f, 0.0f); 
	if(m_reflective_coating_enable) {
		fresnel = m_reflective_coating_90_degrees * 
			(m_reflective_coating_0_degrees + (1.0f - m_reflective_coating_0_degrees) * powf(1.0f - max(0.0f, dot(wi, n)), 5.0f)); 
		if(dot(wi,n) != 0.0f && dot(wo,n) != 0.0f) {
			reflective_coating = m_reflective_coating_reflectance * G(wi, wo, hr, n, m_reflective_coating_roughness) * D(hr, n, m_reflective_coating_roughness) / (4.0f * abs(dot(wi,n) * abs(dot(wo,n))));
		}
	}
	float3 diffuse = make_vector(0.0f, 0.0f, 0.0f); 
	if(diffuse_weight > 0.0f) {
		diffuse = diffuse_weight * m_diffuse_reflectance * (1.0f / M_PI); 
	}
	float3 glossy = make_vector(0.0f, 0.0f, 0.0f); 
	if(glossy_weight > 0.0f) {
		if(dot(wi,n) > 0.0f && dot(wo,n) > 0.0f) {
			glossy = glossy_weight * m_glossy_reflectance * G(wo, wi, hr, n, m_glossy_roughness) * D(hr, n, m_glossy_roughness) / (4.0f * abs(dot(wi,n) * abs(dot(wo,n))));
		}
	}
	float costerm = max(0.0f, dot(wo,n)); 
	return (fresnel * reflective_coating + (1.0f - fresnel) * (diffuse + glossy)) * costerm; 
}

float3 Scene::Material::sample_f(const float3 &i, float3 &o, const Intersection &isect, float &_pdf) const
{
	const float3 &n = isect.facingNormal(i); 
	float3 tangent = normalize(perpendicular(n)); 
	float3 bitangent = cross(n, tangent); 
	// Importance sample only on cosine
	float3 s = cosineSampleHemisphere(); 
	o = normalize(s.x * tangent + s.y * bitangent + s.z * n);
	_pdf = pdf(i,o,isect); 
	return f(i,o,isect);  
};

float Scene::Material::pdf(const float3 &i, const float3 &o, const Intersection &isect) const
{
	const float3 &n = isect.facingNormal(i); 
	float pdf = sameHemisphere(i, o, n) ? dot(o, n) * (1.0f / M_PI) : 0.0f;
	return pdf; 
}

void Scene::ShadowMapShadowGenerator::resize(int width, int height)
{
	m_resolution = make_vector(width, height); 
	if(m_framebuffer_ID == 0) {
		glGenFramebuffers(1, &m_framebuffer_ID); 
		glGenTextures(1, &m_depth_texture); 
	}
	glBindTexture(GL_TEXTURE_2D, m_depth_texture); 
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT16, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
	
	glBindFramebuffer(GL_FRAMEBUFFER, m_framebuffer_ID); 
	glFramebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, m_depth_texture, 0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0); 
}

void Scene::ShadowMapShadowGenerator::update(const vector<Scene::Model *> &models, GLSLProgramObject *program)
{
	view myView = getView(); 
	glPushAttrib(GL_ALL_ATTRIB_BITS); 
	glBindFramebuffer(GL_FRAMEBUFFER, m_framebuffer_ID); 
	glClear(GL_DEPTH_BUFFER_BIT); 
	glViewport(0, 0, m_resolution.x, m_resolution.y); 
	glPolygonOffset(m_polygon_offset_factor, m_polygon_offset_units); 
	glEnable(GL_POLYGON_OFFSET_FILL); 
	program->enable(); 
	myView.push();
	{
		for(auto model : models) {
			// TODO: Check if shadow caster
			model->render(program); 
		}
	}
	myView.pop(); 
	program->disable(); 
	glBindFramebuffer(GL_FRAMEBUFFER, 0); 
	glPopAttrib(); 
}

void Scene::submitShadowGenerators(GLSLProgramObject *programObject)
{
	// This is not pretty. 
	for(int i=0; i<m_lights.size(); i++) {
		stringstream ss; 
		ss << "light_has_shadowmap[" << i << "]"; 
		programObject->setUniform(ss.str().c_str(), false); 
	}


	for(auto shadowGenerator : m_shadowGenerators) {
		for(int i=0; i<m_lights.size(); i++) {
			if(shadowGenerator->m_light == m_lights[i]) {
				shadowGenerator->submit(programObject, i); 
			}
		}
	}
}

void Scene::ShadowMapShadowGenerator::submit(GLSLProgramObject *programObject, int lightID)
{
	{	stringstream ss; 
		ss << "light_has_shadowmap[" << lightID << "]"; 
		programObject->setUniform(ss.str().c_str(), true); 
	}
	{   stringstream ss; 
		ss << "light_shadowmap_sampler[" << lightID << "]"; 
		programObject->bindTexture(ss.str().c_str(), m_depth_texture, GL_TEXTURE_2D, 6 + lightID); 
	}
	{   stringstream ss; 
		ss << "light_shadowmap_mvp[" << lightID << "]"; 
		programObject->setUniform(ss.str().c_str(), getView().get_MVP()); 
	}
}

