#pragma once
#include <utils/OBJModel.h>
#include <utils/view.h>
#include <map>
#include <string>
#include <iostream>
#include <utils/GLSLProgramManager.h>
#include <utils/GLSLProgramObject.h>
#include <utils/Animation.h>
#include <utils/Log.h>
#include <linmath/streamoperators.h>
#include <fstream> 

///////////////////////////////////////////////////////////////////////////////
// Raytracing
///////////////////////////////////////////////////////////////////////////////
#include <embree2/rtcore.h>
#include <embree2/rtcore_ray.h>
#include <linmath/float3.h>
#include "Sampling.h"
inline chag::float3 & asFloat3(float *a) { return *((chag::float3*)a); }
inline chag::float3 & asFloat3(const float *a) { return *((chag::float3*)a); }
inline chag::float3 onRay(const RTCRay &r, float t) {
	return asFloat3(r.org) + t * asFloat3(r.dir); 
}

#pragma warning(disable : 4251)

class Scene; 
class SceneExtender
{
public: 
	virtual bool handleUnknownType(std::string &type, std::istream &is, Scene *scene) = 0; 
};

class Scene
{
public: 
	///////////////////////////////////////////////////////////////////////////
	// The base class for all objects in the scene (models, lights, cameras), 
	// that have an orientation in the world.
	///////////////////////////////////////////////////////////////////////////
	class Item {
		public: 
		virtual ~Item() {}; 
		std::string m_name;
		chag::Orientation m_transform; 
		///////////////////////////////////////////////////////////////////////
		// Any Item must be able to provide a view, so it can be used to view
		// the scene. Items that have a "real" view (e.g. perspective camera 
		// or spot light) override this method. 
		///////////////////////////////////////////////////////////////////////
		virtual view getView() {
			view r; 
			r.pos = m_transform.pos; 
			r.R = m_transform.R; 
			r.m_near = 0.01f; 
			r.m_far = 10000.0f; 
			r.m_fov = 60.0f; 
			r.m_aspect_ratio = 1.0f; 
			return r; 
		}
		void rename(const std::string &new_name) {
			m_name = new_name; 
		};
		virtual Item *clone() = 0; 
	}; 

	///////////////////////////////////////////////////////////////////////////
	// Cameras
	///////////////////////////////////////////////////////////////////////////
	class Camera : public Item {
	public: 
		virtual ~Camera() {}; 
		float m_near; 
		float m_far; 
	};

	class OrthoCamera : public Camera {
		virtual Item *clone() { LOG_ERROR("OrthoCamera::clone not implemented"); return NULL; }; 
	};

	class PerspectiveCamera : public Camera {
	public: 
		float m_fov;
		float m_aspect_ratio; 
		virtual view getView() {
			view r; 
			r.pos = m_transform.pos; 
			r.R = m_transform.R; 
			r.m_near = m_near;
			r.m_far = m_far; 
			r.m_fov = m_fov; 
			r.m_aspect_ratio = m_aspect_ratio; 
			return r;
		}
		virtual Item *clone() {
			PerspectiveCamera *new_item = new PerspectiveCamera; 
			memcpy(new_item, this, sizeof(PerspectiveCamera)); 
			return new_item; 
		}; 
	};

	///////////////////////////////////////////////////////////////////////////
	// Lights
	///////////////////////////////////////////////////////////////////////////
	class Light : public Item {
	public: 
		virtual ~Light() {}; 
		virtual float pdf(const chag::float3 &p, const chag::float3 &lightSample) = 0;
		virtual chag::float3 sample(const chag::float3 &p, float &lightPdf) = 0;
		virtual chag::float3 Le(const chag::float3 &p, const chag::float3 &wo) = 0; 
		virtual bool intersect(const RTCRay &r, float &t) = 0;  
		float m_radiance; 
		chag::float3 m_color;	
	};

	class DiscDiffuseLight : public Light 
	{
	public: 
		float m_radius; 
		float area() { return M_PI * m_radius * m_radius; }; 
		virtual float pdf(const chag::float3 &p, const chag::float3 &lightSample);
		virtual chag::float3 sample(const chag::float3 &p, float &lightPdf);
		virtual chag::float3 Le(const chag::float3 &p, const chag::float3 &wo); 
		virtual bool intersect(const RTCRay &r, float &t);  
		virtual Item *clone() {
			DiscDiffuseLight *new_item = new DiscDiffuseLight; 
			memcpy(new_item, this, sizeof(DiscDiffuseLight)); 
			return new_item; 
		}; 
	};


	///////////////////////////////////////////////////////////////////////////
	// Textures
	// Instead of going all generic here, I will subclass texture for a couple
	// of different types. May or may not be a good idea. Time will tell. 
	///////////////////////////////////////////////////////////////////////////
	class Texture 
	{
	public: 
		std::string m_filename; 
		unsigned int m_width, m_height; 
		GLuint m_GLID; 
		virtual void upload() = 0; 
	};

	class ReflectanceTexture : public Texture
	{
	public: 
		ReflectanceTexture() {
			m_filename = ""; 
			m_width = 0; m_height = 0; m_GLID = 0; m_data = NULL; 
		};
		chag::float3 *m_data; 
		void upload() {}; 
	};

	///////////////////////////////////////////////////////////////////////////
	// Models 
	// Models can be divided into different objects, but their respective 
	// orientation is static. Any model shall be backed by another file. 
	///////////////////////////////////////////////////////////////////////////
	class Intersection; 
	class Material {
	public: 
		std::string m_name; 
		bool			m_reflective_coating_enable; 
		chag::float3	m_reflective_coating_reflectance; 
		float			m_reflective_coating_roughness; 
		float			m_reflective_coating_0_degrees; 
		float			m_reflective_coating_90_degrees; 
		float			m_diffuse_weight; 
		chag::float3	m_diffuse_reflectance; 
		float			m_glossy_weight; 
		chag::float3	m_glossy_reflectance; 
		float			m_glossy_roughness;

		Texture			*m_diffuse_reflectance_texture; 

		Material() {
			m_reflective_coating_enable = false; 
			m_reflective_coating_reflectance = chag::make_vector(0.7f, 0.7f, 0.7f); 
			m_reflective_coating_roughness = 0.1f; 
			m_reflective_coating_0_degrees = 0.2f; 
			m_reflective_coating_90_degrees = 0.9f; 
			m_diffuse_weight = 1.0f; 
			m_diffuse_reflectance = chag::make_vector(0.7f, 0.7f, 0.7f); 
			m_glossy_weight = 0.0f; 
			m_glossy_reflectance = chag::make_vector(0.7f, 0.7f, 0.7f); 
			m_glossy_roughness = 0.4f; 
			m_diffuse_reflectance_texture = NULL; 
		};

		inline float PCF(float a) const { return (a > 0.0f) ? 1.0f : 0.0f; }
		float D(const chag::float3 &m, const chag::float3 &n, const float alpha) const;
		float G1(const chag::float3 &v, const chag::float3 &m,const chag::float3 &n, const float alpha) const;
		inline float G(const chag::float3 &i, const chag::float3 &o, const chag::float3 &m, const chag::float3 &n, float alpha) const { return G1(i,m,n,alpha)*G1(o,m,n,alpha); }
		chag::float3 f(const chag::float3 &i, const chag::float3 &o, const Intersection &isect) const; 
		chag::float3 sample_f(const chag::float3 &i, chag::float3 &o, const Intersection &isect, float &_pdf) const;
		float pdf(const chag::float3 &i, const chag::float3 &o, const Intersection &isect) const; 
	};

	///////////////////////////////////////////////////////////////////////////
	// Helper for buffer objects
	///////////////////////////////////////////////////////////////////////////
	template<class T> 
	class buffer {
	public: 
		std::vector<T>	m_host_vector; 
		GLuint			m_buffer_object;
		buffer() : m_buffer_object(0) {}; 
		void upload() {
			if(m_buffer_object == 0) glGenBuffers(1, &m_buffer_object); 
			glBindBuffer(GL_ARRAY_BUFFER, m_buffer_object); 
			glBufferData(GL_ARRAY_BUFFER, m_host_vector.size() * sizeof(T), &m_host_vector[0], GL_STATIC_DRAW); 
		}; 
	};

	///////////////////////////////////////////////////////////////////////////
	// Model
	///////////////////////////////////////////////////////////////////////////
	class Model : public Item {
		public: 
		~Model() { 
			LOG_ERROR("Scene::Model destructor not implemented (won't clear gfx mem or materials)");
		}; 
		std::string m_filename; 
		std::map<std::string, Material *> m_materials; 
		unsigned int m_embreeGeometryID; 
		buffer<chag::float3> m_positions; 
		buffer<chag::float3> m_normals; 
		buffer<chag::float3> m_tangents; 
		buffer<chag::float2> m_uvs; 
		std::vector<Material *> m_triangleMaterial; 
		std::map<Material *, std::vector<GLint>> m_drawCallOffsets;
		std::map<Material *, std::vector<GLsizei>> m_drawCallCounts; 
		GLuint m_vaob; 
		virtual void render(GLSLProgramObject *programObject); 
		virtual chag::Aabb getAabb(); 
		virtual Item *clone() {
			///////////////////////////////////////////////////////////////////
			// WARNING: With the other objects, a clone is a complete copy,
			//			but here geometry is not copied, only transform and  
			//          material. 
			///////////////////////////////////////////////////////////////////
			Model *new_item = new Model; 
			*new_item = *this; 
			for(auto m : m_materials) {
				Material *material_copy = new Material; 
				*material_copy = *m.second; 
				new_item->m_drawCallCounts[material_copy] = m_drawCallCounts[m.second]; 
				new_item->m_drawCallOffsets[material_copy] = m_drawCallOffsets[m.second]; 
				new_item->m_materials[m.first] = material_copy; 
			}
			return new_item; 
		}; 		
	};

	void loadOBJMaterials(const std::string &filename, const std::string &basePath, Model *model);
	Model * loadOBJModel(std::string &filename, bool importMaterial = false);

	///////////////////////////////////////////////////////////////////////////
	// Shadow Generators
	///////////////////////////////////////////////////////////////////////////

	GLSLProgramManager *m_programManager; 
	GLSLProgramObject  *m_shadowMapProgram; 

	class ShadowGenerator : public Item {
	public: 
		Light *m_light; 
		virtual void update(const std::vector<Scene::Model *> &models, GLSLProgramObject *program) {}; 
		virtual void submit(GLSLProgramObject *programObject, int lightID) {}; 
		virtual void save(std::ofstream &os) = 0; 
	}; 

	class ShadowMapShadowGenerator : public ShadowGenerator {
	public: 
		float m_fov;
		float m_aspect_ratio; 
		float m_near; 
		float m_far; 
		float m_polygon_offset_factor, m_polygon_offset_units;
		chag::int2 m_resolution; 
		GLuint m_depth_texture; 
		GLuint m_framebuffer_ID; 
		ShadowMapShadowGenerator() { m_framebuffer_ID = 0; }; 
		void resize(int width, int height); 
		virtual Item *clone() {
			LOG_WARNING("Clone not yet implemented for ShadowMapShadowGenerator!"); 
			return NULL; 
		}
		virtual void update(const std::vector<Scene::Model *> &models, GLSLProgramObject *program); 
		virtual void submit(GLSLProgramObject *programObject, int lightID); 
		virtual view getView() {
			view r; 
			r.pos = m_light->m_transform.pos; 
			r.R = m_light->m_transform.R; 
			r.m_near = m_near; 
			r.m_far = m_far; 
			r.m_fov = m_fov; 
			r.m_aspect_ratio = m_aspect_ratio; 
			return r; 
		}
		virtual void save(std::ofstream &os) {
			os << "shadowmapshadowgenerator " << m_name << " " << m_light->m_name << " " << m_fov << " " << m_aspect_ratio << " " 
				<< m_near << " " << m_far << " " << m_polygon_offset_factor << " " << m_polygon_offset_units << " " 
				<< m_resolution.x << " " << m_resolution.y << std::endl; 
		}
	};

	void updateShadowGenerators() {
		for(auto sg : m_shadowGenerators) {
			sg->update(m_models, m_shadowMapProgram); 
		}
	}

	void submitShadowGenerators(GLSLProgramObject *programObject);

	///////////////////////////////////////////////////////////////////////////
	// Containers
	///////////////////////////////////////////////////////////////////////////
	std::vector<Model *>	m_models;
	std::vector<Camera *>	m_cameras; 
	std::vector<Light *>	m_lights; 
	std::vector<ShadowGenerator *> m_shadowGenerators; 
	std::map<std::string, Texture *> m_textures; 

	///////////////////////////////////////////////////////////////////////////
	// Environment map
	///////////////////////////////////////////////////////////////////////////
	class EnvironmentMap
	{
	public: 
		EnvironmentMap() { for(int i=0; i<6; i++) m_faces[i] = NULL; m_active = false; }; 
		enum { POSX, POSY, POSZ, NEGX, NEGY, NEGZ }; 
		chag::float3 *m_faces[6]; 
		bool m_active; 
		unsigned int m_width, m_height; 
		GLuint m_texture_id; 
		void load(unsigned char *posx, unsigned char *posy, unsigned char *posz,
				  unsigned char *negx, unsigned char *negy, unsigned char *negz,
				  unsigned int width, unsigned int height); 
	} m_environmentMap;

	///////////////////////////////////////////////////////////////////////////
	// Raytracing
	///////////////////////////////////////////////////////////////////////////
	RTCScene m_embreeScene; 
	std::map<unsigned int, Model *> m_embreeIDToModelMap; 
	class Intersection
	{
	public:
		chag::float3 m_position; 
		chag::float3 m_normal; 
		chag::float3 m_geometryNormal; 
		chag::float3 facingNormal(const chag::float3 &wi) const {
			return (dot(wi, m_normal) > 0.0f) ? m_normal : -m_normal; 
		}; 
		chag::float2 m_uv; 
		Material *m_material;
	};
	void initializeRaytracer(); 
	bool intersectP(RTCRay &r); 
	bool intersect(RTCRay &r, Intersection &isect); 

	///////////////////////////////////////////////////////////////////////////
	// Getters/setters
	///////////////////////////////////////////////////////////////////////////
	Light * getLight(std::string name) {
		for(int i=0; i<m_lights.size(); i++) {
			if(m_lights[i]->m_name == name) return m_lights[i];
		}
		return NULL; 
	}
	Camera * getCamera(std::string name) {
		for(int i=0; i<m_cameras.size(); i++) {
			if(m_cameras[i]->m_name == name) return m_cameras[i];
		}
		return NULL; 
	}
	Model * getModel(std::string name) {
		for(int i=0; i<m_models.size(); i++) {
			if(m_models[i]->m_name == name) return m_models[i];
		}
		return NULL; 
	}

	///////////////////////////////////////////////////////////////////////////
	// Load/Save Scene
	///////////////////////////////////////////////////////////////////////////
	std::string m_filename; 
	bool save(std::string filename);
	bool load(std::string filename, SceneExtender * scene_extender = NULL); 
	bool importOBJ(std::string filename);

	///////////////////////////////////////////////////////////////////////////
	// Geometry
	///////////////////////////////////////////////////////////////////////////
	float m_sceneUnitMultiplier; 
	chag::Aabb getAabb(); 

	///////////////////////////////////////////////////////////////////////////
	// Render
	///////////////////////////////////////////////////////////////////////////
	void submitModels(GLSLProgramObject *programObject); 
	void submitLightGeometry(GLSLProgramObject *programObject); 
	void submitLights(GLSLProgramObject *programObject, chag::orientation &transform); 

	///////////////////////////////////////////////////////////////////////////
	// Camera/Light control
	// The "current controllable" view is the one that gets transformed by 
	// mouse input and what not. The "current renderable" is the one we look
	// from. 
	///////////////////////////////////////////////////////////////////////////
	int m_controllableViewCounter; 
	int m_renderableViewCounter; 
	Item *getCurrentControllableItem();
	void setCurrentControllableItem(const std::string &name); 
	void nextControllableItem();
	void prevControllableItem();
	Item * getCurrentRenderableItem();
	void setCurrentRenderableItem(const std::string &name); 
	void nextRenderableItem();
	void prevRenderableItem();

	Scene(void);
	~Scene(void);
};

inline std::ostream& operator << (std::ostream &os, Scene::Camera &c)
{
	Scene::PerspectiveCamera *pcamera = dynamic_cast<Scene::PerspectiveCamera *>(&c); 
	Scene::OrthoCamera *ocamera = dynamic_cast<Scene::OrthoCamera *>(&c); 
	if(pcamera) {
		os	<< "perspective_camera " << pcamera->m_name << " " << pcamera->m_near << " " << pcamera->m_far << " " 
			<< pcamera->m_fov << " " << pcamera->m_aspect_ratio << " " << pcamera->m_transform; 
	}
	else if(ocamera) {
		os	<< "ortho_camera " << ocamera->m_name << " " << ocamera->m_near << " " << ocamera->m_far << " " 
			<< ocamera->m_transform; 
	}
	else {
		LOG_ERROR("Trying to save unimplemented camera " << c.m_name << "!");
	}
	return os; 
}

inline std::istream& operator >> (std::istream& is, Scene::PerspectiveCamera& c)
{
	is	>> c.m_name >> c.m_near >> c.m_far >> c.m_fov >> c.m_aspect_ratio >> c.m_transform; 
	return is; 
}

inline std::ostream& operator << (std::ostream &os, Scene::Light &c)
{
	Scene::DiscDiffuseLight *ddlight = dynamic_cast<Scene::DiscDiffuseLight *>(&c); 
	if(ddlight) {
		os	<< "disc_diffuse_light " << ddlight->m_name << " " << ddlight->m_color << " " 
			<< ddlight->m_radiance << " " << ddlight->m_radius << " " << ddlight->m_transform; 
	}
	else {
		LOG_ERROR("Trying to save unimplemented light " << c.m_name << "!");
	}
	return os; 
}

inline std::istream& operator >> (std::istream& is, Scene::DiscDiffuseLight &l)
{
	is	>> l.m_name >> l.m_color >> l.m_radiance >> l.m_radius >> l.m_transform; 
	return is; 
}

inline std::ostream& operator << (std::ostream &os, Scene::Model &m)
{
	os << "model " << m.m_name << " " << m.m_filename << " " << m.m_transform; 
	return os; 
}

inline std::istream& operator >> (std::istream& is, Scene::Model &m)
{
	is	>> m.m_name >> m.m_filename >> m.m_transform; 	
	return is; 
}

inline std::ostream& operator << (std::ostream &os, Scene::Material &m)
{
	os	<< m.m_name << " " << (m.m_reflective_coating_enable ? "true" : "false") << " " << m.m_reflective_coating_reflectance << " " 
		<< m.m_reflective_coating_roughness << " " << m.m_reflective_coating_0_degrees << " "
		<< m.m_reflective_coating_90_degrees << " " 
		<< m.m_diffuse_weight << " " << m.m_diffuse_reflectance << " " 
		<< m.m_glossy_weight << " " << m.m_glossy_reflectance << " " << m.m_glossy_roughness; 
	return os; 
}

inline std::istream& operator >> (std::istream &is, Scene::Material &m)
{
	std::string boolval; 
	is >> boolval; 
	m.m_reflective_coating_enable = boolval == "true" ? true : false; 
	is	>> m.m_reflective_coating_reflectance 
		>> m.m_reflective_coating_roughness >> m.m_reflective_coating_0_degrees 
		>> m.m_reflective_coating_90_degrees 
		>> m.m_diffuse_weight >> m.m_diffuse_reflectance
		>> m.m_glossy_weight >> m.m_glossy_reflectance >> m.m_glossy_roughness; 
	return is; 
}
