#include "Scene.h"
#include <sstream>
using namespace std; 
using namespace chag; 

///////////////////////////////////////////////////////////////////////////////
// Helper class to tokenize the obj file input.
///////////////////////////////////////////////////////////////////////////////
struct ObjTri {
	int v[3];
	int t[3];
	int n[3];
	Scene::Material *material; 
};

#define FORCE_INLINE __forceinline
class ObjLexer {
public:
	enum { s_bufferLength = 512 };
	ObjLexer(istream* input) : m_input(input), m_bufferPos(0), m_bufferEnd(0) {}
	enum Tokens	{
		T_Eof      = 0,
		T_MtlLib   = 'm' << 8 | 't',
		T_UseMtl   = 'u' << 8 | 's',
		T_Face     = 'f' << 8 | ' ', // line starts with 'f' followed by ' '
		T_Face2    = 'f' << 8 | '\t', // line starts with 'f' followed by '\t'
		T_Vertex   = 'v' << 8 | ' ', // line starts with 'v' followed by ' '
		T_Vertex2  = 'v' << 8 | '\t', // line starts with 'v' followed by '\t'
		T_Normal   = 'v' << 8 | 'n',
		T_TexCoord = 'v' << 8 | 't',
	};
	FORCE_INLINE int fillBuffer() {
		if (m_bufferPos >= m_bufferEnd)	{
			m_input->read(m_buffer, s_bufferLength);
			m_bufferEnd = int(m_input->gcount());
			m_bufferPos = 0;
		}
		return m_bufferEnd != 0;
	}
	FORCE_INLINE int nextChar()	{
		if (fillBuffer()) {
			return m_buffer[m_bufferPos++];
		}
		return 0;
	}
	int firstLine()	{
		// read the first line token.
		return nextChar() << 8 | nextChar();
	}
	FORCE_INLINE int nextLine()	{
		// scan to end of line...
		while('\n' != nextChar()) {}
		while(matchChar('\n') || matchChar('\r')) {}
		// Or: convert next 2 chars to token (16 bit), can be mt, us, vn, vt, v , v\t, f , f\t, unknown
		return nextChar() << 8 | nextChar();
	}
	FORCE_INLINE bool match(const char s[], const size_t l)	{
		for (int i = 0; fillBuffer() && i < int(l) - 1; ++i) {
			if (s[i] != m_buffer[m_bufferPos]) { return false; }
			else{ ++m_bufferPos; }
		}
		return true;
	}
	FORCE_INLINE bool matchString(std::string &str)	{
		while (fillBuffer() && !isspace(m_buffer[m_bufferPos]))	{
			str.push_back(m_buffer[m_bufferPos++]);
		}
		return !str.empty();
	}
	FORCE_INLINE bool matchFloat(float &result) {
		bool found = false;
		result = 0.0f;
		float sign = 1.0f;
		if (matchChar('-'))	{
			sign = -1.0f;
			found = true;
		}
		char c;
		while (fillBuffer() && myIsDigit(c = m_buffer[m_bufferPos])) {
			result = result * 10.0f + float(c - '0'); 
			++m_bufferPos;
			found = true;
		}
		float frac = 0.1f;
		if (matchChar('.'))	{
			char c;
			while (fillBuffer() && myIsDigit(c = m_buffer[m_bufferPos])) {
				result += frac * float(c - '0'); 
				++m_bufferPos;
				frac *= 0.1f;
			}
			found = true;
		}
		if (matchChar('e') || matchChar('E')) {
			float expSign = matchChar('-') ? -1.0f : 1.0f;
			int exp = 0;
			if (matchInt(exp)) {
				result = result * powf(10.0f, float(exp) * expSign);
			}
		}
		result *= sign;
		return found;
	}
	FORCE_INLINE bool myIsDigit(char c)	{
		return ((unsigned int)(c) - (unsigned int)('0') < 10U);
	}
	FORCE_INLINE bool matchInt(int &result) {
		bool found = false;
		result = 0;
		char c;
		while (fillBuffer() && myIsDigit(c = m_buffer[m_bufferPos])) {
			result = result * 10 + int(c - '0'); 
			++m_bufferPos;
			found = true;
		}
		return found;
	}
	FORCE_INLINE bool matchChar(int matchTo) {
		if (fillBuffer() && m_buffer[m_bufferPos] == matchTo) {
			m_bufferPos++;
			return true;
		}
		return false;
	}
	FORCE_INLINE bool matchWs(bool optional = false) {
		bool found = false;
		while (fillBuffer() && (m_buffer[m_bufferPos] == ' ' || m_buffer[m_bufferPos] == '\t')) {
			found = true;
			m_bufferPos++;
		}
		return found || optional;
	}
	istream* m_input;
	char m_buffer[s_bufferLength];
	int m_bufferPos;
	int m_bufferEnd;
};

FORCE_INLINE static bool parseFaceIndSet(ObjLexer &lexer, ObjTri &t, int v) {
	t.v[v] = -1;
	t.t[v] = -1;
	t.n[v] = -1;
	if(lexer.matchWs(true) 
		&& lexer.matchInt(t.v[v])
		&&   (lexer.matchChar('/') || true)
		&&   (lexer.matchInt(t.t[v]) || true)  // The middle index is optional!
		&&   (lexer.matchChar('/') || true)   // The middle index is optional!
		&&   (lexer.matchInt(t.n[v]) || true ))
	{
		// need to adjust for silly obj 1 based indexing
		t.v[v] -= 1;
		t.t[v] -= 1;
		t.n[v] -= 1;
		return true;
	}
	return false;
}

void Scene::loadOBJMaterials(const string &filename, const string &basePath, Scene::Model *model)
{
	ifstream file;
	file.open(filename.c_str());
	if (!file) {
		LOG_ERROR("Scene::loadOBJMaterials::Error in openening file '" << filename << "'");
		return;
	}
	Material * current_material = NULL;
	std::string lineread;
	while(std::getline(file, lineread)) {
		stringstream ss(lineread); 
		std::string firstword; 
		ss >> firstword;
		std::transform(firstword.begin(), firstword.end(), firstword.begin(), ::tolower); 
		if(firstword == "newmtl") {	
			current_material = new Material; 
			ss >> current_material->m_name;	
			model->m_materials[current_material->m_name] = current_material; 
		}
		else if(firstword == "kd" || firstword == "diffusereflectance") {
			ss	>> current_material->m_diffuse_reflectance.x 
				>> current_material->m_diffuse_reflectance.y 
				>> current_material->m_diffuse_reflectance.z; 
		}
		else if(firstword == "map_kd") {
			ReflectanceTexture *txt = new ReflectanceTexture; 
			ss >> txt->m_filename; 
			m_textures[txt->m_filename] = txt; 
			current_material->m_diffuse_reflectance_texture = txt; 
			LOG_INFO("Has diffuse reflectance texture " << txt->m_filename); 
		}
	}
	return;
}

Scene::Model * Scene::loadOBJModel(string &filename, bool importMaterial)
{
	auto getBasePath = [] (const std::string &filePath) {
		size_t pos = filePath.find_last_of("/\\");
		if (pos != std::string::npos) return filePath.substr(0, pos + 1);
		return std::string();
	};

	Model *model = new Model; 
	model->m_filename = filename; 
	Material *current_material = NULL; 
	std::vector<ObjTri> tris; 
	std::vector<float3> positions; 
	std::vector<float3> normals; 
	std::vector<float2> uvs; 
	ifstream file(filename); 
	if(!file.is_open()) {
		LOG_ERROR("Scene::loadOBJModel::Failed to open " << filename);
		return NULL; 
	}
	ObjLexer lexer(&file);

	for(int token = lexer.firstLine(); token != ObjLexer::T_Eof; token = lexer.nextLine())	{
		switch(token) {
		case ObjLexer::T_MtlLib: {
			string materialFile;
			if(lexer.match("llib", sizeof("llib")) && lexer.matchWs() && lexer.matchString(materialFile)) {
				if(importMaterial)
					loadOBJMaterials(getBasePath(filename) + materialFile, getBasePath(filename), model);
			}
			break;
		}
		case ObjLexer::T_UseMtl: {
			string materialName;
			if(lexer.match("emtl", sizeof("emtl")) && lexer.matchWs() && lexer.matchString(materialName)) {
				if(importMaterial) {
					current_material = model->m_materials[materialName];
				}
				else {
					// If we are not importing materials, these will be overwritten later. 
					if(model->m_materials[materialName] != NULL) {
						current_material = model->m_materials[materialName];
					}
					else {
						current_material = new Material; 
						model->m_materials[materialName] = current_material; 
					}
				}
			}
			break;
		}
		case ObjLexer::T_Vertex:
		case ObjLexer::T_Vertex2:
			{
				float3 p;
				if (lexer.matchWs(true)
					&& lexer.matchFloat(p.x)
					&& lexer.matchWs()
					&& lexer.matchFloat(p.y)
					&& lexer.matchWs()
					&& lexer.matchFloat(p.z))
				{
					positions.push_back(p);
				}
			}
			break;
		case ObjLexer::T_Normal: 
			{
				float3 n = { 0.0f, 0.0f, 0.0f };
				if (lexer.matchWs(true)
					&& lexer.matchFloat(n.x)
					&& lexer.matchWs()
					&& lexer.matchFloat(n.y)
					&& lexer.matchWs()
					&& lexer.matchFloat(n.z))
				{
					normals.push_back(n);
				}
			}
			break;
		case ObjLexer::T_TexCoord: 
			{
				float2 t;
				if (lexer.matchWs(true)
					&& lexer.matchFloat(t.x)
					&& lexer.matchWs()
					&& lexer.matchFloat(t.y))
				{
					uvs.push_back(t);
				}
			}
			break;
		case ObjLexer::T_Face: 
		case ObjLexer::T_Face2:
			{
				ObjTri t;
				// parse vert 0 and 1
				if (parseFaceIndSet(lexer, t, 0) && parseFaceIndSet(lexer, t, 1))
				{
					// let any more produce one more triangle
					while(parseFaceIndSet(lexer, t, 2))
					{
						// Drop degenerate triangles
						float3 v0 = positions[t.v[0]];
						float3 v1 = positions[t.v[1]];
						float3 v2 = positions[t.v[2]];
						const float epsilon = 0.0000001f; 
						if(length(v0-v1) < epsilon || length(v1-v2) < epsilon || length(v2-v0) < epsilon) {
							//degenerate_triangles += 1; 
						}
						else {
							// kick tri,
							t.material = current_material;
							tris.push_back(t);
							// the make last vert second (this also keeps winding the same).
							t.n[1] = t.n[2];
							t.t[1] = t.t[2];
							t.v[1] = t.v[2];
						}
					}
				}
			}
			break;
		default:
			break;
		};
	}

	current_material = NULL; 
	int offset = 0; 
	int count = 0; 
	for(auto triangle : tris) {
		for(int i=0; i<3; i++) {
			if(triangle.material != current_material) {
				if(current_material != NULL)
					model->m_drawCallCounts[current_material].push_back(count); 
				count = 0; 
				current_material = triangle.material; 
				model->m_drawCallOffsets[current_material].push_back(offset); 
			}
			model->m_positions.m_host_vector.push_back(positions[triangle.v[i]]);
			model->m_normals.m_host_vector.push_back(normals[triangle.n[i]]);
			// TODO: Fix tangents!!!!
			float3 tangent = perpendicular(normals[triangle.n[i]]); 
			model->m_tangents.m_host_vector.push_back(tangent);
			if(triangle.t[i] != -1)
				model->m_uvs.m_host_vector.push_back(uvs[triangle.t[i]]);
			else 
				model->m_uvs.m_host_vector.push_back(make_vector(0.0f, 0.0f));
			offset += 1;
			count += 1; 
		}
		// Used for the raytracing code
		model->m_triangleMaterial.push_back(triangle.material); 
	}
	model->m_drawCallCounts[current_material].push_back(count); 

	glGenVertexArrays(1, &model->m_vaob); 
	glBindVertexArray(model->m_vaob); 
	model->m_positions.upload(); 
	glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0); 
	glEnableVertexAttribArray(0);
	model->m_normals.upload(); 
	glVertexAttribPointer(1, 3, GL_FLOAT, false, 0, 0); 
	glEnableVertexAttribArray(1);
	model->m_tangents.upload(); 
	glVertexAttribPointer(2, 3, GL_FLOAT, false, 0, 0); 
	glEnableVertexAttribArray(2);
	model->m_uvs.upload(); 
	glVertexAttribPointer(3, 2, GL_FLOAT, false, 0, 0); 
	glEnableVertexAttribArray(3);

	return model;
}
