#include "DebugRenderer.h"
#include "ChagGlUtils.h"
#include "Assert.h"
#include <GL/glew.h>
#include <linmath/int2.h>
#include <linmath/int3.h>
#include <linmath/float3x3.h>
#include <linmath/float4x4.h>
#include <linmath/float3.h>
#include <linmath/float2.h>
#include <GL/freeglut.h>


static void drawCone(const chag::float3 &axis, const float angle, const chag::float3 &pos, bool drawAxis = true)
{
	if (drawAxis)
	{
		glLineWidth(3.0f);
		glBegin(GL_LINES);
			glVertex(pos);
			glVertex(pos + axis * 1.3f);
		glEnd();
		glLineWidth(1.0f);
	}
	glPushMatrix();
	glMultMatrix(make_matrix_from_zAxis(pos + axis, -axis, perp(axis)));
	float height = chag::length(axis);
	glutSolidCone(tanf(angle) * height, height, 48, 2);
	glPopMatrix();

}


static void drawAabb(const chag::Aabb &aabb)
{
  using namespace chag;

  chag::float3 mi = aabb.min;
  chag::float3 ma = aabb.max;

  // face 0 min X
  glVertex(mi);
  glVertex(make_vector(mi.x, ma.y, mi.z));
  glVertex(make_vector(ma.x, ma.y, mi.z));
  glVertex(make_vector(ma.x, mi.y, mi.z));
  // face 1, max X
  glVertex(make_vector(mi.x, mi.y, ma.z));
  glVertex(make_vector(ma.x, mi.y, ma.z));
  glVertex(ma);
  glVertex(make_vector(mi.x, ma.y, ma.z));

  // face 0 min Y
  glVertex(mi);
  glVertex(make_vector(ma.x, mi.y, mi.z));
  glVertex(make_vector(ma.x, mi.y, ma.z));
  glVertex(make_vector(mi.x, mi.y, ma.z));
  // face 1, max Y
  glVertex(make_vector(mi.x, ma.y, mi.z));
  glVertex(make_vector(mi.x, ma.y, ma.z));
  glVertex(ma);
  glVertex(make_vector(ma.x, ma.y, mi.z));

  // face 0 min Z
  glVertex(mi);
  glVertex(make_vector(mi.x, mi.y, ma.z));
  glVertex(make_vector(mi.x, ma.y, ma.z));
  glVertex(make_vector(mi.x, ma.y, mi.z));
  // face 1, max Z      
  glVertex(make_vector(ma.x, mi.y, mi.z));
  glVertex(make_vector(ma.x, ma.y, mi.z));
  glVertex(ma);
  glVertex(make_vector(ma.x, mi.y, ma.z));
}



namespace chag
{

const chag::float4 DebugRenderer::red    = {1.0f, 0.0f, 0.0f, 1.0f};
const chag::float4 DebugRenderer::blue   = {0.0f, 0.0f, 1.0f, 1.0f};
const chag::float4 DebugRenderer::green  = {0.0f, 1.0f, 0.0f, 1.0f};
const chag::float4 DebugRenderer::yellow = {1.0f, 1.0f, 0.1f, 1.0f};
const chag::float4 DebugRenderer::purple = {0.4f, 0.1f, 0.4f, 1.0f};
const chag::float4 DebugRenderer::cyan   = {0.0f, 1.0f, 1.0f, 1.0f};
const chag::float4 DebugRenderer::orange = {1.0f, 0.5f, 0.0f, 1.0f};
const chag::float4 DebugRenderer::brown  = {0.55f, 0.27f, 0.07f, 1.0f};
const chag::float4 DebugRenderer::pink   = {0.78f, 0.44f, 0.44f, 1.0f};
const chag::float4 DebugRenderer::olive  = {0.56f, 0.56f, 0.22f, 1.0f};

DebugRenderer& DebugRenderer::instance()
{
  static DebugRenderer inst;
  return inst;
}

chag::float4 DebugRenderer::randomColor(uint32_t i, float alpha)
{
	static const chag::float4 colors[] =
	{
		red,
		green,
		blue,
		purple,
		yellow,
		cyan,
		orange,
		brown, 
		pink,
		olive,
	};
	float4 c = colors[i % (sizeof(colors) / sizeof(colors[0]))];
	c.w = alpha;
	return c;
}

void DebugRenderer::addCone(const float3 &pos, const float3 &dir, float height, float angle, const float4 &color, const std::string &layer)
{
	Item::Cone cone = { pos, height, dir, angle };
	Item item;
	item.kind = Item::K_Cone;
	item.color = color;
	item.cone = cone;
	m_items[layer].items.push_back(item);
}



void DebugRenderer::addAabb(const chag::Aabb &aabb, const float4 &color, const std::string &layer)
{
	Item item;
	item.kind = Item::K_Aabb;
	item.color = color;
	item.aabb = aabb;
	m_items[layer].items.push_back(item);
}



void DebugRenderer::addSphere(const float3 &pos, float rad, const float4 &color, const std::string &layer)
{
	Item::Sphere sphere = { pos, rad };
	Item item;
	item.kind = Item::K_Sphere;
	item.color = color;
	item.sphere = sphere;
	m_items[layer].items.push_back(item);
}

void DebugRenderer::addLine(const chag::float3 &pos0, const chag::float3 &pos1, const chag::float4 &color, const std::string &layer)
{
	Item::Line line = {pos0, pos1};
	Item item; 
	item.kind = Item::K_Line;
	item.color = color; 
	item.line = line; 
	m_items[layer].items.push_back(item);
}

void DebugRenderer::addTriangle(const chag::float3 &pos0, const chag::float3 &pos1, const chag::float3 &pos2, const chag::float4 &color, const std::string &layer)
{
	Item::Triangle triangle = {pos0, pos1, pos2};
	Item item; 
	item.kind = Item::K_Triangle;
	item.color = color; 
	item.triangle = triangle; 
	m_items[layer].items.push_back(item);
}

void DebugRenderer::clear(const std::string &layer, const float4x4 &layerTransform)
{
	if (layer == std::string())
	{
		m_items.clear();
	}
	else
	{
		Layer &l = m_items[layer];
		l.items.clear();
		l.transform = layerTransform;
	}
}





void DebugRenderer::flush(const std::string &layer)
{
	if (layer == std::string())
	{
		//flushLayer()
	}
	else
	{
		flushLayer(layer);
	}
}


void DebugRenderer::flushLayer(const std::string &layerId)
{
	const Layer &layer = m_items[layerId];
	const std::vector<Item> &items = layer.items;

	glPushMatrix();
	glMultMatrix(layer.transform);

	for (std::vector<Item>::const_iterator it = items.begin(); it != items.end(); ++it)
	{
		const Item &item = (*it);
		glColor(item.color);
		switch(item.kind)
		{
		case Item::K_Cone:
			{
				const Item::Cone &cone = item.cone;
				drawCone(cone.direction * cone.height, cone.angle, cone.position, false);
			};
			break;
		case Item::K_Sphere:
			{
				const Item::Sphere &sphere = item.sphere;
				glPushMatrix();
				glTranslate(sphere.position);
				glutSolidSphere(sphere.radius, 6, 6);
				glPopMatrix();				
			};
			break;
		case Item::K_Aabb:
			{
				const Aabb &aabb = item.aabb;
				glBegin(GL_QUADS);
				drawAabb(aabb);
				glEnd();
			};
			break;
		case Item::K_Line: 
			{
				glBegin(GL_LINES); 
				glVertex3fv(&item.line.pos0.x); 
				glVertex3fv(&item.line.pos1.x); 
				glEnd(); 
			}
			break;
		case Item::K_Triangle: 
			{
				glBegin(GL_TRIANGLES); 
				glVertex(item.triangle.pos0); 
				glVertex(item.triangle.pos1); 
				glVertex(item.triangle.pos2); 
				glEnd(); 
			}
			break;
		};
	}

	glPopMatrix();
}


DebugRenderer::DebugRenderer()
{
}


}; // namespace chag

