#include "DeviceMemoryTracker.h"
#include "Assert.h"

#if ENABLE_DEVICE_MEMORY_TRACKER

static size_t internalFormatSize(GLint internalFormat)
{
  switch(internalFormat)
  {
  case GL_RGBA32F_ARB:
    return 16;
  case GL_RGB32F_ARB:
    return 12;
  case GL_ALPHA32F_ARB:
    return 4;
  case GL_INTENSITY32F_ARB:
    return 4;
  case GL_LUMINANCE32F_ARB:
    return 4;
  case GL_LUMINANCE_ALPHA32F_ARB:
    return 8;
  case GL_RGBA16F_ARB:
    return 8;
  case GL_RGB16F_ARB:
    return 6;
  case GL_ALPHA16F_ARB:
    return 2;
  case GL_INTENSITY16F_ARB:
    return 2;
  case GL_LUMINANCE16F_ARB:
    return 2;
  case GL_LUMINANCE_ALPHA16F_ARB:
    return 4;
  case GL_ALPHA:
  case GL_ALPHA8:
  case GL_LUMINANCE:
  case GL_LUMINANCE8:
  case GL_INTENSITY:
  case GL_INTENSITY8:
    return 1;
  case GL_ALPHA16:
  case GL_LUMINANCE16:
  case GL_INTENSITY16:
  case GL_LUMINANCE_ALPHA:
  case GL_LUMINANCE8_ALPHA8:
  case GL_LUMINANCE12_ALPHA4:
    return 2;
  case GL_LUMINANCE16_ALPHA16:
    return 4;
  case GL_R3_G3_B2:
    return 1;
  case GL_RGB:
    return 3;
  case GL_RGB5:
    return 2;
  case GL_RGB8:
    return 3;
  case GL_RGB10:
    return 4;
  case GL_RGB16:
    return 6;
  case GL_RGBA:
    return 4;
  case GL_RGBA2:
    return 1;
  case GL_RGBA4:
    return 2;
  case GL_RGB5_A1:
    return 2;
  case GL_RGBA8:
    return 4;
  case GL_RGB10_A2:
    return 4;
  case GL_RGBA16:
    return 8;
  case GL_UNSIGNED_INT_24_8_EXT:
    return 4;
  case GL_DEPTH24_STENCIL8_EXT:
  case GL_DEPTH_COMPONENT32:
  case GL_DEPTH_COMPONENT32F_NV:
  case GL_DEPTH_COMPONENT32F:
    return 4;
  case GL_LUMINANCE4:
  case GL_LUMINANCE12:
  case GL_LUMINANCE6_ALPHA2:
  case GL_LUMINANCE4_ALPHA4:
  case GL_LUMINANCE12_ALPHA12:
  case GL_INTENSITY4:
  case GL_INTENSITY12:
  case GL_ALPHA4:
  case GL_ALPHA12:
  case GL_RGB4:
  case GL_RGBA12:
  case GL_RGB12:
  case GL_TEXTURE_RED_TYPE_ARB:
  case GL_TEXTURE_GREEN_TYPE_ARB:
  case GL_TEXTURE_BLUE_TYPE_ARB:
  case GL_TEXTURE_ALPHA_TYPE_ARB:
  case GL_TEXTURE_LUMINANCE_TYPE_ARB:
  case GL_TEXTURE_INTENSITY_TYPE_ARB:
  case GL_TEXTURE_DEPTH_TYPE_ARB:
  case GL_UNSIGNED_NORMALIZED_ARB:
  default:
    break;
  };

  // unknown format
  DBG_BREAK();

  return 0;
}



DeviceMemoryTracker& DeviceMemoryTracker::instance()
{
  static DeviceMemoryTracker inst;
  return inst;
}



cudaError_t DeviceMemoryTracker::cudaMalloc(void **devPtr, size_t size)
{
  cudaError_t result = ::cudaMalloc(devPtr, size);

  m_totalCuda += size;
  m_cudaAllocs[*devPtr] = AllocRecord(size, m_currentLine, m_currentFile);

  return result;
}



cudaError_t DeviceMemoryTracker::cudaFree(void *devPtr)
{
  std::map<void *, AllocRecord>::iterator it = m_cudaAllocs.find(devPtr);
  if (it != m_cudaAllocs.end())
  {
    m_totalCuda -= (*it).second.m_size;
    m_cudaAllocs.erase(it);
  }

  return ::cudaFree(devPtr);
}



void DeviceMemoryTracker::glTexImage2D(GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
  GLint currentTexture = 0;
  glGetIntegerv(GL_TEXTURE_2D_BINDING_EXT, &currentTexture);

  m_totalGl -= m_textureAllocs[currentTexture].m_size;
  size_t byteSize = width * height * internalFormatSize(internalformat);
  m_totalGl += byteSize;
  m_textureAllocs[currentTexture] = AllocRecord(byteSize, m_currentLine, m_currentFile);

  ::glTexImage2D(target, level, internalformat, width, height, border, format, type, pixels);
}



void DeviceMemoryTracker::glBufferData(GLenum target, GLsizeiptr size, const GLvoid* data, GLenum usage)
{
  GLint currentBuffer = 0;
  glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &currentBuffer);

  m_totalGl -= m_bufferAllocs[currentBuffer].m_size;
  m_totalGl += size;
  m_bufferAllocs[currentBuffer] = AllocRecord(size, m_currentLine, m_currentFile);

  ::glBufferData(target, size, data, usage);
}



void DeviceMemoryTracker::glDeleteTextures(GLsizei n, const GLuint *textures)
{
  for (GLsizei i = 0; i < n; ++i)
  {
    std::map<GLuint, AllocRecord>::iterator it = m_textureAllocs.find(textures[i]);
    if (it != m_textureAllocs.end())
    {
      m_totalGl -= (*it).second.m_size;
      m_textureAllocs.erase(it);
    }
  }
  ::glDeleteTextures(n, textures);
}



void DeviceMemoryTracker::glDeleteBuffers(GLsizei n, const GLuint* buffers)
{
  for (GLsizei i = 0; i < n; ++i)
  {
    std::map<GLuint, AllocRecord>::iterator it = m_bufferAllocs.find(buffers[i]);
    if (it != m_bufferAllocs.end())
    {
      m_totalGl -= (*it).second.m_size;
      m_bufferAllocs.erase(it);
    }
  }
  ::glDeleteBuffers(n, buffers);
}



void DeviceMemoryTracker::setLineAndFile(int line, const char *file)
{
  m_currentLine = line;
  m_currentFile = file;
}


DeviceMemoryTracker::DeviceMemoryTracker() : 
  m_totalCuda(0),
  m_totalGl(0),
  m_currentFile(__FILE__),
  m_currentLine(0)
{
}


#endif // ENABLE_DEVICE_MEMORY_TRACKER
