#ifndef _DeviceMemoryTracker_h_
#define _DeviceMemoryTracker_h_

#define ENABLE_DEVICE_MEMORY_TRACKER 0
#if ENABLE_DEVICE_MEMORY_TRACKER
#include "IntTypes.h"
#include <GL/glew.h>
#include <GL/glut.h>
#include <map>
#include <cuda_runtime_api.h>

/**
 * Use to track memory usage, should be used to track GL and CUDA allocations
 */
class DeviceMemoryTracker
{
public:

  static DeviceMemoryTracker& instance();
  
  // cuda wrappers.
  cudaError_t cudaMalloc(void **devPtr, size_t size);
  cudaError_t cudaFree(void *devPtr);

  // gl wrappers
  void glTexImage2D(GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
  void glBufferData(GLenum target, GLsizeiptr size, const GLvoid* data, GLenum usage);
  void glDeleteTextures(GLsizei n, const GLuint *textures);
  void glDeleteBuffers(GLsizei n, const GLuint* buffers);

  void setLineAndFile(int line, const char *file);

  size_t getTotalAllocated() const { return m_totalCuda + m_totalGl; }
  size_t getAllocatedCuda() const { return m_totalCuda; }
  size_t getAllocatedGl() const { return m_totalGl; }

private:

  DeviceMemoryTracker();

  struct AllocRecord
  {
    explicit AllocRecord(size_t size = 0, int line = 0, const char *file = 0) : 
      m_size(size), 
      m_line(line), 
      m_file(file) 
    { 
    }

    size_t m_size;
    int m_line;
    const char *m_file;
  };

  std::map<GLuint, AllocRecord> m_textureAllocs;
  std::map<GLuint, AllocRecord> m_bufferAllocs;
  std::map<void *, AllocRecord> m_cudaAllocs;

  size_t m_totalCuda;
  size_t m_totalGl;

  const char *m_currentFile;
  int m_currentLine;
};

#define TRACK_MEM(_call_with_args_) (DeviceMemoryTracker::instance().setLineAndFile(__LINE__, __FILE__), DeviceMemoryTracker::instance()._call_with_args_)

#else // !ENABLE_DEVICE_MEMORY_TRACKER

#define TRACK_MEM(_call_with_args_) _call_with_args_

#endif // ENABLE_DEVICE_MEMORY_TRACKER

#endif // _DeviceMemoryTracker_h_
