#include <cstdlib>			// exit()
#include <utils/Log.h>
#include "GLSLProgramManager.h"
#define NV_REPORT_COMPILE_ERRORS
//#include "./nvShaderUtils.h"

using namespace std; 

GLSLProgramManager::GLSLProgramManager(string filename)
{
	setFile(filename);
}

void GLSLProgramManager::setFile(string filename)
{
	m_fullSource.clear();
	LOG_INFO("Loading shader collection: " << filename);
	// Load the file as a list of strings
	m_filename = filename; 
	ifstream f;
	f.open(filename.c_str());
	if(!f.is_open()){
		LOG_ERROR("Loading shader collection: " << filename << "FAILED.");
	}
	string lineread;
	while(std::getline(f, lineread))
	{
#ifndef WINDOWS
		if (lineread.size() > 0 && lineread.end()[-1] == '\r')
			lineread.erase(lineread.end()-1);		// strip DOS line ending
#endif
		m_fullSource.push_back(lineread); 
	}
}

GLSLProgramManager::~GLSLProgramManager(void)
{
}

string GLSLProgramManager::extractShader(string shaderName)
{
	// First find what common clauses to use, if any
	vector<string> is_using; 
	for(unsigned int i=0; i<m_fullSource.size(); i++){
		string &line = m_fullSource[i];
		if(line == string("shader ") + shaderName) {			
			string line2 = m_fullSource[i+1];
			if(line2.substr(0, 5) == "using"){
				size_t current = 6; 
				size_t next = -1; 
				do
				{
					next = line2.find(" \0", current+1);
					if(next == -1) next = line2.size(); 
					string common = line2.substr(current, next - current); 
					current = next + 1; 
					is_using.push_back(common);
				}
				while(current < line2.size());
			}
		}
	}

	// Create the source files without losing line information
	string shader; 
	// Add the defines
	for(map<string, string>::iterator i = m_defines.begin(); i!=m_defines.end(); i++){
		shader += "#define "; 
		shader += (*i).first;
		shader += " ";
		shader += (*i).second; 
		shader += "\n";
	}

	bool shaderFound = false; 
	bool commonFound = false; 
	for(unsigned int i=0; i<m_fullSource.size(); i++){
		string line = m_fullSource[i]; 
		if(line == string("shader ") + shaderName) {
			shaderFound = true; 
			char buffer[100];
			sprintf_s(buffer, "#line %d\n", i+1);
			shader += buffer; 
		}
		else if(line.substr(0,6) == "common"){
			string block = line.substr(7, line.size());
			for(unsigned int j=0; j<is_using.size(); j++){
				if(block == is_using[j]) {
					commonFound = true; 
					char buffer[100];
					sprintf_s(buffer, "#line %d\n", i+1);
					shader += buffer; 
				}
			}
		}
		else if(line.substr(0,1) == "{"){
			if(commonFound || shaderFound) {
				shader += "\n"; 
			}
		}
		else if(line.substr(0,1) == "}"){
			if(shaderFound) break;
			if(commonFound) {
				shader += "\n";
				commonFound = false; 
			}
		}
		else if(commonFound) shader += line + "\n";
		else if(shaderFound) {
			if(line.substr(0, 5) == "using") line = string("//") + line;
			else shader += line + "\n"; 
		}
	}
	if(!shaderFound) return "";
	return shader; 
}

void GLSLProgramManager::compileProgram(const string &vertex_shader, 
										const string &fragment_shader,
										GLSLProgramObject *program)
{
	LOG_INFO("Creating GLSL program: (" << vertex_shader << ", " << fragment_shader << ") from " << m_filename);
	string vertexShaderSource = extractShader(vertex_shader);
	if(vertexShaderSource == "") {
		LOG_ERROR("Failed to extract shader \"" << vertex_shader);
		exit(1); 
	}
	string fragmentShaderSource = extractShader(fragment_shader);
	if(fragmentShaderSource == "") {
		LOG_ERROR("Failed to extract shader \"" << fragment_shader);
		exit(1); 
	}
	program->compileProgram(vertexShaderSource.c_str(), NULL, fragmentShaderSource.c_str());
}

void GLSLProgramManager::compileProgram(const string &vertex_shader, 
										const string &geometry_shader, 
										const string &fragment_shader,
								        GLenum gsInput, 
										GLenum gsOutput,
										unsigned int verticesOut,
										GLSLProgramObject *program)
{
	LOG_INFO("Creating GLSL program: (" << vertex_shader << ", " << geometry_shader << ", " << fragment_shader << ") from " << m_filename);
	string vertexShaderSource = extractShader(vertex_shader);
	if(vertexShaderSource == "") {
		LOG_ERROR("Failed to extract shader \"" << vertex_shader);
		exit(1); 
	}
	string geometryShaderSource = extractShader(geometry_shader);
	if(geometryShaderSource == "") {
		LOG_ERROR("Failed to extract shader \"" << geometry_shader);
		exit(1); 
	}
	string fragmentShaderSource = extractShader(fragment_shader);
	if(fragmentShaderSource == "") {
		LOG_ERROR("Failed to extract shader \"" << fragment_shader);
		exit(1); 
	}
	program->compileProgram(vertexShaderSource.c_str(), geometryShaderSource.c_str(), fragmentShaderSource.c_str(), 
							gsInput, gsOutput, verticesOut);
}

GLSLProgramObject * GLSLProgramManager::getProgramObject(const string &vertex_shader, 
														 const string &fragment_shader)
{
	GLSLProgramObject *program = new GLSLProgramObject;
	compileProgram(vertex_shader, fragment_shader, program);
	LoadedProgram thisOne = {vertex_shader, "", fragment_shader, 0, 0, 0, program};
	m_loadedPrograms.push_back(thisOne);
	return program; 
}

GLSLProgramObject * GLSLProgramManager::getProgramObject(const string &vertex_shader, 
														 const string &geometry_shader, 
														 const string &fragment_shader,
								                         GLenum gsInput, 
														 GLenum gsOutput,
														 unsigned int verticesOut)
{
	GLSLProgramObject *program = new GLSLProgramObject;
	compileProgram(vertex_shader, geometry_shader, fragment_shader, gsInput, gsOutput, verticesOut, program);
	LoadedProgram thisOne = {vertex_shader, geometry_shader, fragment_shader, gsInput, gsOutput, verticesOut, program};
	m_loadedPrograms.push_back(thisOne);
	return program; 
}


// Re-read, re-extract and re-compile shaders.
// Useful if m_defines have been changed, or if file has been edited.
void GLSLProgramManager::reload()
{
	setFile(m_filename);
	for(unsigned int i=0; i<m_loadedPrograms.size(); i++)
	{
		m_loadedPrograms[i].programObject->destroy();
		if(m_loadedPrograms[i].geometryShader == ""){
			compileProgram(m_loadedPrograms[i].vertexShader, 
						   m_loadedPrograms[i].fragmentShader, 
						   m_loadedPrograms[i].programObject);
		}
		else {
			compileProgram(	m_loadedPrograms[i].vertexShader,
							m_loadedPrograms[i].geometryShader,
							m_loadedPrograms[i].fragmentShader,
							m_loadedPrograms[i].geomIn, 
							m_loadedPrograms[i].geomOut, 
							m_loadedPrograms[i].verticesIn, 
							m_loadedPrograms[i].programObject);
		}
	}
}

void GLSLProgramManager::setPreprocDef(const char *name, int value)
{
  char buf[11];
  sprintf(buf, "%d", value);
  m_defines[name] = buf;
}



void GLSLProgramManager::setPreprocDef(const char *name, bool value)
{
  m_defines[name] = value ? "1" : "0";
}
