#include <stdlib.h>
#include <iostream>
#include <string>
using namespace std; 
#include "GLSLProgramObject.h"
#include "Log.h"

GLSLProgramObject::GLSLProgramObject(const char *vsource, const char *fsource)
{
    mProg = compileProgram(vsource, 0, fsource);
}

GLSLProgramObject::GLSLProgramObject(const char *vsource, const char *gsource, const char *fsource,
                         GLenum gsInput, GLenum gsOutput)
{
    mProg = compileProgram(vsource, gsource, fsource, gsInput, gsOutput);
}

GLSLProgramObject::~GLSLProgramObject()
{
	if (mProg) {
		glDeleteProgram(mProg);
	}
}

void GLSLProgramObject::destroy()
{
	if (mProg) {
		glDeleteProgram(mProg);
	}
}

void
GLSLProgramObject::enable()
{
	glUseProgram(mProg);
}

void
GLSLProgramObject::disable()
{
	glUseProgram(0);
}

int GLSLProgramObject::getLocation(const char *name){
#if _DEBUG
	if(glGetUniformLocation(mProg, name) < 0){
        //LOG_WARNING("Error setting parameter " << name);
	}
#endif
	return glGetUniformLocation(mProg, name); 
}

///////////////////////////////////////////////////////////////////////////////
// Set one uniform
///////////////////////////////////////////////////////////////////////////////
void GLSLProgramObject::setUniform(const char *name, const float &value){
	glUniform1f(getLocation(name), value); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::float2 &value){
	glUniform2f(getLocation(name), value.x, value.y); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::float3 &value){
	glUniform3f(getLocation(name), value.x, value.y, value.z); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::float4 &value){
	glUniform4f(getLocation(name), value.x, value.y, value.z, value.w); 
}
void GLSLProgramObject::setUniform(const char *name, const int &value){
	glUniform1i(getLocation(name), value); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::int2 &value){
	glUniform2i(getLocation(name), value.x, value.y); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::int3 &value){
	glUniform3i(getLocation(name), value.x, value.y, value.z); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::int4 &value){
	glUniform4i(getLocation(name), value.x, value.y, value.z, value.w); 
}
void GLSLProgramObject::setUniform(const char *name, const unsigned int &value){
	glUniform1ui(getLocation(name), value); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::float4x4 &value){
	glUniformMatrix4fv(getLocation(name), 1, false, &value.c1.x);
}
void GLSLProgramObject::setUniform(const char *name, const chag::float3x3 &value){
	glUniformMatrix3fv(getLocation(name), 1, false, &value.c1.x);
}

///////////////////////////////////////////////////////////////////////////////
// Set uniform array
///////////////////////////////////////////////////////////////////////////////
void GLSLProgramObject::setUniform(const char *name, const float *value, int count){
	glUniform1fv(getLocation(name), count, value); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::float2 *value, int count){
	glUniform2fv(getLocation(name), count, &(value->x)); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::float3 *value, int count){
	glUniform3fv(getLocation(name), count, &(value->x));
}
void GLSLProgramObject::setUniform(const char *name, const chag::float4 *value, int count){
	glUniform4fv(getLocation(name), count, &(value->x));
}
void GLSLProgramObject::setUniform(const char *name, const int *value, int count){
	glUniform1iv(getLocation(name), count, value); 
}
void GLSLProgramObject::setUniform(const char *name, const chag::int2 *value, int count){
	glUniform2iv(getLocation(name), count, &(value->x));
}
void GLSLProgramObject::setUniform(const char *name, const chag::int3 *value, int count){
	glUniform3iv(getLocation(name), count, &(value->x));
}
void GLSLProgramObject::setUniform(const char *name, const chag::int4 *value, int count){
	glUniform4iv(getLocation(name), count, &(value->x));
}
void GLSLProgramObject::setUniform(const char *name, const chag::float4x4 *value, int count){
	glUniformMatrix4fv(getLocation(name), count, false, &value[0].c1.x);
}

void GLSLProgramObject::bindTexture(const char *name, GLuint tex, GLenum target, GLint unit)
{
    GLint loc = getLocation(name); 
    glActiveTexture(GL_TEXTURE0 + unit);
    glBindTexture(target, tex);
    glUseProgram(mProg);
    glUniform1i(loc, unit);
    glActiveTexture(GL_TEXTURE0);
}


GLuint GLSLProgramObject::compileShader(const char *source, GLenum shaderType)
{
	GLuint shader = glCreateShader(shaderType); 
	glShaderSource(shader, 1, &source, 0);
	glCompileShader(shader); 
	GLint status; 
	glGetShaderiv(shader, GL_COMPILE_STATUS, &status); 
	if(status == GL_FALSE){
		int loglength; 
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &loglength); 
		char *log = new char[loglength];
		int returnedLength; 
		glGetShaderInfoLog(shader, loglength, &returnedLength, log); 
		string shaderTypeName;
		if(shaderType == GL_VERTEX_SHADER) shaderTypeName = "vertex";
		if(shaderType == GL_FRAGMENT_SHADER) shaderTypeName = "fragment";
		if(shaderType == GL_GEOMETRY_SHADER) shaderTypeName = "geometry";
		
		LOG_INFO(shaderTypeName << " shader failed to compile:\n" << log);
		delete[] log; 
	}
	return shader; 
}

GLuint GLSLProgramObject::compileProgram(const char *vsource, const char *gsource, const char *fsource,
                            GLenum gsInput, GLenum gsOutput, unsigned int verticesOut)
{
	GLuint program = glCreateProgram();
    glAttachShader(program, compileShader(vsource, GL_VERTEX_SHADER));
    glAttachShader(program, compileShader(fsource, GL_FRAGMENT_SHADER));
    if (gsource) {
		GLuint geomShader = glCreateShader(GL_GEOMETRY_SHADER_ARB);
        glShaderSource(geomShader, 1, &gsource, 0);
        glCompileShader(geomShader);
        glAttachShader(program, geomShader);

        glProgramParameteriEXT(program, GL_GEOMETRY_INPUT_TYPE_EXT, gsInput);
        glProgramParameteriEXT(program, GL_GEOMETRY_OUTPUT_TYPE_EXT, gsOutput); 
        glProgramParameteriEXT(program, GL_GEOMETRY_VERTICES_OUT_EXT, verticesOut); 

    }
    glLinkProgram(program);

    // check if program linked
    GLint success = 0;
    glGetProgramiv(program, GL_LINK_STATUS, &success);
    if (!success) {
        char temp[4096];
        glGetProgramInfoLog(program, 4096, 0, temp);
        LOG_ERROR("Failed to link program:\n" << temp);
        glDeleteProgram(program);
        program = 0;
        exit(-1);
    }

	mProg = program; 
    return program;
}
